// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Arrangement/include/CGAL/Arr_segment_traits_2.h,v $
// $Revision: 1.20 $ $Date: 2003/09/18 10:19:41 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Iddo Hanniel <hanniel@math.tau.ac.il>
//                 Efi Fogel    <efif@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_ARR_SEGMENT_EXACT_TRAITS_T2_H
#define CGAL_ARR_SEGMENT_EXACT_TRAITS_T2_H

#include <CGAL/Pm_segment_traits_T2.h>
#include <CGAL/intersections.h>

#include <list>

CGAL_BEGIN_NAMESPACE

template <class Kernel_>
class Arr_segment_traits_T2 : public Pm_segment_traits_T2<Kernel_>
{
protected:
  typedef typename Kernel::rep                             rep;

public:
  typedef Kernel_                                          Kernel;
  typedef int                                              Info_face;
  typedef int                                              Info_edge;
  typedef int                                              Info_vertex;

  typedef Pm_segment_traits_T2<Kernel>                     Base;

  typedef typename Base::Has_left_category                 Has_left_category;

  typedef typename Base::Point_T2                          Point_T2;
  typedef typename Base::X_monotone_curve_T2               X_monotone_curve_T2;
  typedef typename Base::Intersect_T2                      Intersect_T2;
  typedef typename Base::Cmp_dist_collinear_origin_T2      Cmp_dist_collinear_origin_T2;
  typedef typename Base::Line_T2                           Line_T2;
  typedef typename Base::Object_T2                         Object_T2;

  typedef X_monotone_curve_T2                              Curve_T2;

  // Obsolete, for backward compatibility
  typedef Point_T2                                         Point;
  typedef X_monotone_curve_T2                              X_curve;
  typedef Curve_T2                                         Curve;

  typedef typename Kernel::Construct_vertex_T2             Construct_vertex_T2;
  typedef typename Kernel::Equal_T2                        Equal_T2;
  typedef typename Kernel::Orientation_T2                  Orientation_T2;
  typedef typename Kernel::Collinear_has_on_T2             Collinear_has_on_T2;
  typedef typename Kernel::Antipode_T2                     Antipode_T2;
  typedef typename Kernel::Construct_opposite_segment_T2   Construct_opposite_segment_T2;
  typedef typename Kernel::Collinear_T2                    Collinear_T2;
typedef typename rep::Point_3                            Point_R3;


public:
  Arr_segment_traits_T2() : Base() { }

  /*! is_x_monotone()
   * \return true if the given curve is an x-monotone curve. False, otherwise.
   * For segments, this is always true
   */
  bool is_dirSweepLine_monotone(const Curve_T2 &) {return true;}

  bool is_defined_left(const Point_T2 & p, const Curve_T2 &curve) const
  {
    Point_T2 source = curve_point(curve,0);
    Point_T2 target = curve_point(curve,1);
    Point_T2 o(1,0,0);
    Point_T2 rel = source;
    if (point_equal(p, source) )
      rel = target;
      
    return (orientation(o, p, rel) < 0);
  }

 
  bool has_on(const Curve_T2 &c, const Point_T2 &p) 
  { return c.has_on(p); }
  
  /*! curve_make_x_monotone() cuts the given curve into x-monotone subcurves
   * and inserts them to the given output iterator. The order in which they
   * are inserted defines their order in the hierarchy tree.
   * While segments are x_monotone, still need to pass them out.
   * \param cv the input curve
   * \param o the output iterator
   * \return the past-the-end iterator
   */
  template<class OutputIterator>
  OutputIterator curve_make_dirSweepLine_monotone(const Curve_T2 & cv,
                                       OutputIterator o) const
  {
    *o++ = cv;
    return o;
  } 

  bool compare_around_origin(const Point_T2 & p1,
                             const Point_T2 & p2) const
  {
    return compare_around_origin_T2_object()(p1,p2); 
  }


  /*! curve_opposite() flips a given curve
   * \param cv the curve
   * \return a segment with source and target point interchanged
   */
  X_monotone_curve_T2 curve_opposite(const X_monotone_curve_T2 & cv) const
  {
    return construct_opposite_segment_T2_object()(cv);
  }
 
  
  /*! nearest_intersection_to_right() finds the nearest intersection point of
   * two given curves to the right of a given point. Nearest is defined as the
   * lexicographically nearest not including the point itself with one
   * exception explained bellow..
   * If the intersection of the two curves is an X_monotone_curve_T2, that is,
   * there is an overlapping subcurve, then if the the source and target of the
   * subcurve are strickly to the right, they are returned through two
   * other point references p1 and p2. If pt is between the source and target
   * of the overlapping subcurve, or pt is its left endpoint, pt and the target
   * of the right endpoint of the subcurve are returned through p1 and p2 
   * respectively.
   * If the intersection of the two curves is a point to the right of pt, pt
   * is returned through the p1 and p2.
   * \param c1 the first curve
   * \param c2 the second curve
   * \param pt the point to compare against
   * \param p1 the first point reference
   * \param p2 the second point reference
   * \return true if c1 and c2 do intersect to the right of pt. Otherwise,
   * false
   */
  bool nearest_intersection_to_right(const X_monotone_curve_T2 & c1,
                                     const X_monotone_curve_T2 & c2,
                                     const Point_T2 & pt,
                                     Point_T2 & p1, Point_T2 & p2) const
  {
  
    typedef typename Point_T2::FT FT;
    Object res = intersect_T2_object()(c1, c2);
    Point_T2 o(FT(1),FT(0),FT(0));
  
    // Empty object is returned - no intersection.
    if (res.is_empty()) return (false);

    // Intersection is a point
    if (assign(p1,res)) {
      // the intersection is a point:
  
      if (orientation_T2_object()(o,pt,p1) == 1) {
  
        p2 = p1;
        return true;
      }
  
      return false;
    }
    
    // Intersection is a segment
    X_monotone_curve_T2 seg;
    if (assign(seg, res)) {
  
      // the intersection is a curve:
      Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
  
      const Point_T2 & src = construct_vertex(seg, 0);
  
      const Point_T2 & trg = construct_vertex(seg, 1);
  
      const Point_T2 o(FT(1),FT(0),FT(0));
  
      Orientation o1 = orientation(o, pt, src);
  
      Orientation o2 = orientation(o, pt, trg);

      if (o1==1 && o2==1) {
        // the subcurve is completely to the right:
        p1 = src;
        p2 = trg;
        return true;
      }
      
      if (o1==1) {
        // target is to the left, source is to the right:
        p1 = src;
        p2 = pt;
        return true;
      }

      if (o2==1) {
        // source is to the left, target is to the right:
        p1 = pt;
        p2 = trg;
        return true;
      }

      // the subcurve is completely to the left:
      return false;
    }

    // the curves do not intersect:
    return false;
  }

  
  
  bool collinear_has_on(const Point_T2 p,
                        const Point_T2 q,
                        const Point_T2 r) const 
  {
    return collinear_has_on_T2_object()(p,q,r);
  }

  Point_T2 antipode(const Point_T2 p) const 
  {
    return antipode_T2_object()(p);
  }

  Object_T2 intersect(const X_monotone_curve_T2 &cv1,
                      const X_monotone_curve_T2 &cv2)	
  {
    return intersect_T2_object()(cv1,cv2);
  }

  bool subdivide_curve(X_monotone_curve_T2 & curve_base,
                       X_monotone_curve_T2 &subcurve1,
                       X_monotone_curve_T2 &subcurve2)
  {
    typedef typename Point_T2::FT FT;
    const Point_T2 p(FT(0),FT(1),FT(0));
    Intersect_T2 intersect = intersect_T2_object();
    Antipode_T2 antipode = antipode_T2_object();
    const Point_T2 origin(FT(1),FT(0),FT(0));
    if (orientation(origin,curve_base.source(),curve_base.target()) == 0) {
      Collinear_has_on_T2 collinear_has_on = collinear_has_on_T2_object();
      if (collinear_has_on(origin,curve_base.source(),curve_base.target())) {
        subcurve1 = Curve_T2(curve_base.source(),origin);
        subcurve2 = Curve_T2(origin,curve_base.target());
        return true;
      }
      if (collinear_has_on(antipode(origin),curve_base.source(),curve_base.target())) {
        subcurve1 = Curve_T2(curve_base.source(),antipode(origin));
        subcurve2 = Curve_T2(antipode(origin),curve_base.target());
        return true;
      }
    }
    Object_T2 o = intersect(curve_base,Curve_T2(origin,p));
    Point_T2 point;
    if (assign(point,o)) {
      subcurve1 = Curve_T2(curve_base.source(),point);
      subcurve2 = Curve_T2(point,curve_base.target());
      return true;
    }
    else {
      o = intersect(curve_base,Curve_T2(p,antipode(origin)));
      if (assign(point,o)) {
        subcurve1 = Curve_T2(curve_base.source(),point);
        subcurve2 = Curve_T2(point,curve_base.target());
        return true;
      }
    }
    return false;
  }

  /*! curves_overlap() test overlapping between two given curves
   * \patam c1 the first curve
   * \patam c2 the second curve
   * \return true if c1 and c2 overlap in a one-dimensional subcurve
   * (i.e., not in a finite number of points). Otherwise, false.
   */
  bool curves_overlap(const X_monotone_curve_T2 & cv1,
                      const X_monotone_curve_T2 & cv2) const
  {
    Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
    const Point_T2 & src2 = construct_vertex(cv2, 0);
    const Point_T2 & trg2 = construct_vertex(cv2, 1);
    const Point_T2 & src1 = construct_vertex(cv1, 0);
    const Point_T2 & trg1 = construct_vertex(cv1, 1);

    Orientation_T2 orient = orientation_T2_object();
    if ((!orient(src1, trg1, src2) == COLLINEAR) ||
        (!orient(src1, trg1, trg2) == COLLINEAR))
      return false;
    Collinear_has_on_T2 collinear_has_on = collinear_has_on_T2_object();
    if (collinear_has_on(src2,src1,trg1))
      return true;
    return (collinear_has_on(trg2,src1,trg1));
  } 

  Comparison_result compare_distance(const Point_T2 & p1, const Point_T2 & p2,
                                                          const Point_T2 & p3) const
  {
    //if both p1 and p2 are at infinity, they must be the same.
    if (p1.w() == 0) {
      if (p2.w() < 0)
        return SMALLER;
      if (p2.w() > 0)
        return LARGER;
      return EQUAL;
    }
    if (p2.w() == 0) {
      if (p1.w() > 0)
        return SMALLER;
      return LARGER;
    }
    return compare_distance_to_point_T2_object()(p3,p1,p2);	
  }

  Object_T2 intersection (const Curve_T2 &l,
                          const X_monotone_curve_T2 &c) const
  {
    return intersect_T2_object()(l,c);
  }

  Orientation orientation(const Point_T2 &p1,
                          const Point_T2 &p2,
                          const Point_T2 &p3) const
  {
    return orientation_T2_object()(p1,p2,p3);
  }

  Comparison_result compare_right(const X_monotone_curve_T2 &c1, 
                                  const X_monotone_curve_T2 &c2,
                                  const Point_T2 &q)  const 
  {
    typedef typename Point_T2::FT FT;

    Point_T2 o(FT(1),FT(0),FT(0)),p,t;
    //std::cout << "compare_right " << std::endl;
    Object_T2 obj1 = intersect_T2_object()(c1,Line_T2(o,q));
    Object_T2 obj2 = intersect_T2_object()(c2,Line_T2(o,q));
    assign(p,obj1);
    assign(t,obj2);
 
    return (compare_distance(p,t,o));
  }


  Comparison_result cmp_dist_collinear_origin (const Point_T2 & p1,
                                               const Point_T2 & p2) {
    return cmp_dist_collinear_origin_T2_object()(p1,p2);
  }


  bool curves_equal(const X_monotone_curve_T2 & cv1,
                    const X_monotone_curve_T2 & cv2) const
  {
    if (cv1.source() != cv2.source() &&
        cv1.source() != cv2.target())
      return false;
    return (cv1.target() == cv2.source() ||
            cv1.target() == cv2.target());
  }

  Point_T2 curve_point(const X_monotone_curve_T2 & cv, const int i) const
  {
    CGAL_precondition(i == 0 || i == 1);
    if (i == 0)
      return cv.source();
    return cv.target();
  }

  bool is_SourceFirstProcessed(const X_monotone_curve_T2 &  curve,
                               Point_T2 & m_lastPoint) {

    Point_T2 o(1,0,0);
    bool m_isSourceFirstProcessed;
    const Point_T2 & m_source = curve.source();
    const Point_T2 & m_target = curve.target();
    Orientation ori = orientation(o, m_source, m_target);
    if ( ori  == -1 )
    {
      m_lastPoint = m_target; 
  
      m_isSourceFirstProcessed = false;
    }
    else if ( ori == 1)
    {
      m_lastPoint = m_source; 
  
      m_isSourceFirstProcessed = true;
    }
   else {
      if (m_source == antipode(o)) {
        if (!has_on_segment(m_target,o,Point_T2(0,1,0)) && 
                !has_on_segment(m_target,antipode(o),Point_T2(0,1,0))) {
          m_isSourceFirstProcessed = true;
          m_lastPoint = m_source;
        } 
        else {
          m_isSourceFirstProcessed = false;
          m_lastPoint = m_target;
        }
      } 
      else if (m_target == antipode(o)) {
        if (has_on_segment(m_source,o,Point_T2(0,1,0)) ||
                has_on_segment(m_source,antipode(o),Point_T2(0,1,0))){  
          m_isSourceFirstProcessed = true;
          m_lastPoint = m_source;
        } 
        else {
          m_isSourceFirstProcessed = false;
          m_lastPoint = m_target;
        }
      } 
      else if (cmp_dist_collinear_origin(m_source,m_target) == SMALLER) 
      {
        m_isSourceFirstProcessed = true;
        m_lastPoint = m_source;
      }
      else {
        m_isSourceFirstProcessed = false;
        m_lastPoint = m_target;
      }
    }
    return m_isSourceFirstProcessed;
  }
    
};

CGAL_END_NAMESPACE

#endif
