// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Bbox_2.h,v $
// $Revision: 1.12 $ $Date: 2003/10/21 12:18:01 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_BBOX_T2_H
#define CGAL_BBOX_T2_H

#include <CGAL/Segment_T2_Segment_T2_intersection.h>
#include <CGAL/Iso_rectangle_T2.h>
#include <CGAL/Segment_T2.h>

CGAL_BEGIN_NAMESPACE


template <class R_>
class Bbox_T2 : Handle_for<Triple< typename R_::Point_T2, 
																	 typename R_::Point_T1,
																	 typename R_::Point_T1> >
{
  typedef R_ R;
  typedef typename R::FT                        FT;
  typedef typename R::Point_T2                  TPoint_T2;
  typedef typename R::Point_T1                  TPoint_T1;
  typedef typename R::Iso_rectangle_T2          TIso_rectangle_T2;
  typedef typename R::Line_T2                   TLine_T2;
  typedef typename R::Segment_T2                TSegment_T2;
  typedef typename R::Bbox_T2                   TBbox_T2;
  typedef typename R_::Bbox_handle_T2           base;
  typedef typename base::element_type           rep;  
  
private:

  void initialize(const TPoint_T2 &p, const TPoint_T1 &delta_x,
                  const TPoint_T1 &delta_y)
  {
    CGAL_kernel_precondition(!is_at_back_range_T1(delta_x)  &&
                             !is_at_back_range_T1(delta_y));
    if (is_at_infinity(p)) 
      CGAL_kernel_precondition((p==TPoint_T2(0,1,0)) || (p==TPoint_T2(0,0,1)) ||
                               (p==TPoint_T2(0,-1,0)) || (p==TPoint_T2(0,0,-1)));
    initialize_with(rep(p,delta_x,delta_y));
  }

  
public:
  Bbox_T2(){}

  Bbox_T2(const TPoint_T2 &p, const TPoint_T1 &delta_x,
                  const TPoint_T1 &delta_y)
	{
    initialize(p,delta_x,delta_y);
	}

  Bbox_T2(const FT& w0, const FT& x0, const FT& y0, 
          const FT& w1, const FT& x1,
          const FT& w2, const FT& y2)
  {
    TPoint_T2 p(w0, x0, y0);
    TPoint_T1 delta_x(w1, x1);
    TPoint_T1 delta_y(w2, y2);
    initialize(p,delta_x,delta_y);    
  }

  TPoint_T2  p0() const;
  TPoint_T2  p1() const;
  TPoint_T1  delta_x() const;
  TPoint_T1  delta_y() const;
  

  FT  x0() const;
  FT  y0() const;
  FT  w0() const;
  FT  x1() const;
  FT  y1() const;
  FT  w1() const;
  

  TBbox_T2  operator+(const TBbox_T2& b) const;
};

template < class R >
inline
typename R::FT 
Bbox_T2<R>::x0() const
{ return Ptr()->first.x(); }

template < class R >
inline
typename R::FT 
Bbox_T2<R>::y0() const
{ return Ptr()->first.y(); }

template < class R >
inline
typename R::FT 
Bbox_T2<R>::w0() const
{ return Ptr()->first.w(); }

template < class R >
inline
typename R::FT 
Bbox_T2<R>::x1() const
{ return Ptr()->second.x(); }

template < class R >
inline
typename R::FT 
Bbox_T2<R>::y1() const
{ return Ptr()->second.y(); }

template < class R >
inline
typename R::FT 
Bbox_T2<R>::w1() const
{ return Ptr()->second.w(); }

template < class R >
inline
typename Bbox_T2<R>::TPoint_T2  
Bbox_T2<R>::p0() const
{
  return Ptr()->first;
}

template < class R >
inline
typename Bbox_T2<R>::TPoint_T2 
Bbox_T2<R>::p1() const
{
  typename R::FT pw = p0().w()*delta_x().w()*delta_y().w();
  typename R::FT px = delta_y().w()*
                      (p0().x()*delta_x().w() +
                       p0().w()*delta_x().x());
  typename R::FT py = delta_x().w()* 
                      (p0().y()*delta_y().w() + 
                       p0().w()*delta_y().x());
  return TPoint_T2(pw,px,py);
}

template < class R >
inline
typename Bbox_T2<R>::TPoint_T1  
Bbox_T2<R>::delta_x() const
{
  return Ptr()->second;
}

template < class R >
inline
typename Bbox_T2<R>::TPoint_T1 
Bbox_T2<R>::delta_y() const
{
  return Ptr()->third;
}

//#ifndef CGAL_NO_ISTREAM_EXTRACT_BBOX_3
template < class R >
inline
std::istream&
operator>>(std::istream &is, Bbox_T2<R>& b)
{
  typename R::FT w0, x0, y0, w1, x1, w2, y2;

  switch(is.iword(IO::mode))
  {
    case IO::ASCII :
        is >> w0 >> x0 >> y0 >> w1 >> x1 >> w2 >> y2;
        break;
    case IO::BINARY :
        read(is, w0);
        read(is, x0);
        read(is, y0);
        read(is, w1);
        read(is, x1);
        read(is, w2);
        read(is, y2);
        break;
  }
  b = Bbox_T2<R>(w0, x0, y0, w1, x1, w2, y2);
  return is;
}

//#endif // CGAL_NO_ISTREAM_EXTRACT_BBOX_3


//#ifndef CGAL_NO_OSTREAM_INSERT_BBOX_3
template < class R >
inline
std::ostream&
operator<<(std::ostream &os, const Bbox_T2<R>& b)
{
  switch(os.iword(IO::mode))
  {
    case IO::ASCII :
        return os << b.w0() << ' ' << b.x0() << ' ' << b.y0() << 
                  ' ' << b.delta_x().w() << ' ' << b.delta_x().x() << ' ' 
		  << b.delta_y().w() << ' ' << b.delta_y().x();
    case IO::BINARY :
        write(os, b.w0());
        write(os, b.x0());
        write(os, b.y0());
        write(os, b.delta_x().w());
        write(os, b.delta_x().x());
        write(os, b.delta_y().w());
        write(os, b.delta_y().x());
        return os;
    default:
        os << "Bbox_T2<R>((" << b.w0()
           << ", "       << b.x0()
           << ", "       << b.y0() << "), (";
        os <<               b.delta_x().w()
           << ", "       << b.delta_x().x()
           << ", "       << b.delta_y().w() 
           << ", "       << b.delta_y().x() << "))";
        return os;
  }
}
//#endif // CGAL_NO_OSTREAM_INSERT_BBOX_3



template < class R >
inline
typename Bbox_T2<R>::TBbox_T2
Bbox_T2<R>::operator+(const typename Bbox_T2<R>::TBbox_T2 & b) const
{
  if (is_at_infinity(p0()) && is_at_infinity(b.p0())) {
    CGAL_kernel_precondition(p0() == Point_T2<R>(0,1,0) ||
                             p0() == Point_T2<R>(0,-1,0) ||
                             p0() == Point_T2<R>(0,0,1) ||
                             p0() == Point_T2<R>(0,0,-1));
    CGAL_kernel_precondition(p0() == b.p0());
    return Bbox_T2<R>(p0(),delta_x(),delta_y());
  }
  if (is_at_infinity(p0())) {
    CGAL_kernel_precondition(p0() == Point_T2<R>(0,1,0) ||
                             p0() == Point_T2<R>(0,-1,0) ||
                             p0() == Point_T2<R>(0,0,1) ||
                             p0() == Point_T2<R>(0,0,-1));
    if (p0() == Point_T2<R>(0,1,0) ||
        p0() == Point_T2<R>(0,-1,0)) {
      CGAL_kernel_precondition(!is_at_infinityT1(b.delta_y()));
      if (is_at_infinityT1(b.delta_x())) 
        CGAL_kernel_precondition(b.delta_x() == Point_T1<R>(p0().w(),p0().x()));
      return Bbox_T2<R>(b.p0(),Point_T1<R>(p0().w(),p0().x()),b.delta_y());
    }
    if (p0() == Point_T2<R>(0,0,1) ||
        p0() == Point_T2<R>(0,0,-1)) {
      CGAL_kernel_precondition(!is_at_infinityT1(b.delta_x()));
      if (is_at_infinityT1(b.delta_y())) 
        CGAL_kernel_precondition(b.delta_y() == Point_T1<R>(p0().w(),p0().y()));
      return Bbox_T2<R>(b.p0(),b.delta_x(),Point_T1<R>(p0().w(),p0().y()));
    }
  }
  if (is_at_infinity(b.p0())) {
    CGAL_kernel_precondition(b.p0() == Point_T2<R>(0,1,0) ||
                             b.p0() == Point_T2<R>(0,-1,0) ||
                             b.p0() == Point_T2<R>(0,0,1) ||
                             b.p0() == Point_T2<R>(0,0,-1));
    if (b.p0() == Point_T2<R>(0,1,0) ||
        b.p0() == Point_T2<R>(0,-1,0)) {
      CGAL_kernel_precondition(!is_at_infinityT1(delta_y()));
      if (is_at_infinityT1(delta_x())) 
        CGAL_kernel_precondition(delta_x() == Point_T1<R>(b.p0().w(),b.p0().x()));
      return Bbox_T2<R>(p0(),Point_T1<R>(b.p0().w(),b.p0().x()),delta_y());
    }
    if (b.p0() == Point_T2<R>(0,0,1) ||
        b.p0() == Point_T2<R>(0,0,-1)) {
      CGAL_kernel_precondition(!is_at_infinityT1(delta_x()));
      if (is_at_infinityT1(delta_y())) 
        CGAL_kernel_precondition(delta_y() == Point_T1<R>(b.p0().w(),b.p0().y()));
      return Bbox_T2<R>(p0(),delta_x(),Point_T1<R>(b.p0().w(),b.p0().y()));
    }
  }
  std::vector<Point_T2<R> > vpoint,vpt0;
  std::vector<Segment_T2<R> > vSeg;
  std::vector<Segment_T2<R> > hSeg;
  Point_T2<R> pt0, pt1;

  vpoint.push_back(p0());
  vpoint.push_back(b.p0());
  if (p0()!=p1())
    vpoint.push_back(p1());
  if (b.p0()!=b.p1())
    vpoint.push_back(b.p1());
  int i=0;
  int j=0;
  int k=0;

  while(i < vpoint.size() && j < 2 && k < 2)
  {
    if (vpoint[i] != Point_T2<R>(0,0,1) &&
        vpoint[i] != Point_T2<R>(0,0,-1) ) {
      Line_T2<R> l(vpoint[i],Point_T2<R>(0,0,1));
      if (at_the_same_side_or_on(vpoint[(i+1)%vpoint.size()],
                           vpoint[(i+2)%vpoint.size()],
                           vpoint[(i+3)%vpoint.size()],
                           l)) {
        vSeg.push_back(Segment_T2<R>(vpoint[i],Point_T2<R>(0,0,1)));

        j++;
      }
    }
    if (vpoint[i] != Point_T2<R>(0,1,0) &&
        vpoint[i] != Point_T2<R>(0,-1,0) ) {
    Line_T2<R> r(vpoint[i],Point_T2<R>(0,1,0)); 
    if (at_the_same_side_or_on(vpoint[(i+1)%vpoint.size()],
                           vpoint[(i+2)%vpoint.size()],
                           vpoint[(i+3)%vpoint.size()],
                           r)) {
        hSeg.push_back(Segment_T2<R>(vpoint[i],Point_T2<R>(0,1,0)));

        k++;
      }
    }
    i++;
  }
  for (i=0; i < 2; i++) {
    for (j=0; j < 2; j++) {
      Segment_T2_Segment_T2_pair<R> segpair(&vSeg[i],&hSeg[j]);
      segpair.intersection(pt0);
      vpt0.push_back(pt0);
    }
  }
  TPoint_T1 delta_x, delta_y;

  delta_x = scaled_distance_to_line(pt0, Line_T2<R>(pt1,Point_T2<R>(0,0,1)));
  delta_y = scaled_distance_to_line(pt0, Line_T2<R>(pt1,Point_T2<R>(0,1,0)));
 
  i=0;
  j=-1;
  while(i < vpt0.size() && j==-1) {
    if (vpt0[i] == Point_T2<R>(0,0,1) || 
        vpt0[i] == Point_T2<R>(0,0,-1)) {
      delta_x =scaled_distance_to_line(vpt0[(i+1)%vpt0.size()], 
                               Line_T2<R>(vpt0[(i+3)%vpt0.size()],
                               Point_T2<R>(0,0,1)));
      j=(i+1)%vpt0.size();
    }
    if (vpt0[i] == Point_T2<R>(0,1,0) || 
        vpt0[i] == Point_T2<R>(0,-1,0)) {
      delta_x =scaled_distance_to_line(vpt0[(i+1)%vpt0.size()], 
                               Line_T2<R>(vpt0[(i+3)%vpt0.size()],
                               Point_T2<R>(0,1,0)));
      j=(i+1)%vpt0.size();
    }
    i++;
  }
  if (j==-1) j=0;
  if (TPoint_T2(delta_x.w()*vpt0[j].w(),
                vpt0[j].x()*delta_x.w() + vpt0[j].w()*delta_x.x(),
                vpt0[j].y()*delta_x.w()) != vpt0[(j+2)%vpt0.size()])
    delta_x = TPoint_T1(delta_x.w(),-delta_x.x());
  if (TPoint_T2(delta_y.w()*vpt0[j].w(),
                vpt0[j].x()*delta_y.w(),
                vpt0[j].y()*delta_y.w() + vpt0[j].w()*delta_y.x()) != vpt0[(j+2)%vpt0.size()])
    delta_y = TPoint_T1(delta_y.w(),-delta_y.x());
  return Bbox_T2<R>(vpt0[j],delta_x,delta_y);
}   

template < class R >
inline
bool
do_overlap(const Bbox_T2<R>& bb1, const Bbox_T2<R>& bb2)
{
  Iso_rectangle_T2<R> r0(bb1.p0(),bb1.delta_x(),bb1.delta_y());
  Iso_rectangle_T2<R> r1(bb2.p0(),bb2.delta_x(),bb2.delta_y());
  int i = 0;
  int j = 0;
  if (!r0.has_on_unbounded_side(r1[0]))
    return true;
  if (!r0.has_on_unbounded_side(r1[1]))
    return true;
  if (!r0.has_on_unbounded_side(r1[2]))
    return true;
  if (!r0.has_on_unbounded_side(r1[3]))
    return true;
  if (!r1.has_on_unbounded_side(r0[0]))
    return true;
  if (!r1.has_on_unbounded_side(r0[1]))
    return true;
  if (!r1.has_on_unbounded_side(r0[2]))
    return true;
  if (!r1.has_on_unbounded_side(r0[3]))
    return true;
  Segment_T2<R> s0, s1;
  bool overlap = false;
  while (i < 4 && !overlap) {
    while (j < 4 && !overlap) {
      s0 = Segment_T2<R>(r0[i],r0[(i+1)%4]);
      s1 = Segment_T2<R>(r0[j],r0[(j+1)%4]);
      if (do_intersect(s0,s1))
        overlap = true;
      j++;
    }
    i++;
  }
  return overlap;   
}

CGAL_END_NAMESPACE

#endif
