// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Circle_2.h,v $
// $Revision: 1.12 $ $Date: 2003/10/21 12:18:01 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
//                 Sven Schoenherr
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_CIRCLE_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_CIRCLE_T2_H

#include <CGAL/predicates_on_points_T2.h>

CGAL_BEGIN_NAMESPACE

template <class R_ >
class Circle_T2
  : public Handle_for<Triple< typename R_::Point_T2,  typename R_::Point_T2, Orientation> >
{
CGAL_VC7_BUG_PROTECTED
  typedef typename R_::FT                       FT;
  typedef typename R_::Circle_T2                TCircle_T2;
  typedef typename R_::Point_T2                 TPoint_T2;
  typedef typename R_::Point_T1                 TPoint_T1;
  typedef typename R_::Segment_T2               TSegment_T2;
  typedef typename R_::Line_T2                  TLine_T2;
  typedef typename R_::Aff_transformation_T2   TAff_transformation_2;

  typedef typename R_::Circle_handle_T2          base;
  typedef typename base::element_type            rep;

public:
  typedef R_                                     R;

  Circle_T2()
    : base() {}

  
  Circle_T2(const TPoint_T2 &center, const Orientation &orient)
  {
    CGAL_kernel_precondition( orient != COLLINEAR );

    initialize_with(rep(center, center, orient));
  }

  Circle_T2(const TPoint_T2 &center, const TPoint_T2 &p,
            const Orientation &orient = COUNTERCLOCKWISE) 
  { 
    CGAL_kernel_precondition(!antipodal(p,center));
    CGAL_kernel_precondition(orient != COLLINEAR);
    initialize_with(rep(center,p,orient));
  }

 
  bool           operator==(const TCircle_T2 &s) const;
  bool           operator!=(const TCircle_T2 &s) const;
 
  
  const TPoint_T1 & squared_radius() const
  {
    TPoint_T1 *sr;
    if (is_at_infinity(center()) || is_at_infinity(point()))
    {
      sr = new TPoint_T1(0,1);
      return *sr;
    }
    sr = new TPoint_T1(squared_distance(center(),point()));
    return *sr;
  }

  const Orientation orientation() const
  {
    return Ptr()->third;
  }
  
  const TPoint_T2 & center() const
  {
    return Ptr()->first;
  }

  const TPoint_T2 & point() const
  {
    return Ptr()->second;
  }

  TCircle_T2  opposite() const;


  Oriented_side  oriented_side(const TPoint_T2 &p) const;
  Bounded_side   bounded_side(const TPoint_T2 &p) const;

  bool has_on_boundary(const TPoint_T2 &p) const;
  bool has_on_negative_side(const TPoint_T2 &p) const;
  bool has_on_positive_side(const TPoint_T2 &p) const;

  bool has_on_bounded_side(const TPoint_T2 &p) const;
  bool has_on_unbounded_side(const TPoint_T2 &p) const;

  bool is_degenerate() const;

};


template < class R >
CGAL_KERNEL_INLINE
bool
Circle_T2<R>::operator==(const TCircle_T2 &c) const
{ 
  if (identical(c))
    return true;
  return center() == c.center() &&
         point() == c.point() &&
         orientation() == c.orientation();
}

template < class R >
inline
bool
Circle_T2<R>::operator!=(const TCircle_T2 &c) const
{
  return !(*this == c);
}

template < class R >
inline
bool
Circle_T2<R>::is_degenerate() const
{
  return center()==point();
}


template < class R >
inline
typename Circle_T2<R>::TCircle_T2
Circle_T2<R>::opposite() const
{
  return Circle_T2<R>(center(),point(),CGAL::opposite(orientation()) );
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
Oriented_side
Circle_T2<R>::
oriented_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return Oriented_side(bounded_side(p) * orientation());
}

template < class R >
CGAL_KERNEL_INLINE
Bounded_side
Circle_T2<R>::
bounded_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  TPoint_T1 r = squared_radius();
  if (is_at_infinity(p)) {
    if (center().w()*p.w() > 0)
      return ON_BOUNDED_SIDE;
    if (p.w() == 0)
      return ON_BOUNDARY;
    return ON_UNBOUNDED_SIDE;
  }
  if (is_at_infinity(center())) {
   //compara distancia relativa de p e de point() a center
   Comparison_result cr;
   if ((cr = compare_relative_distance_to_omega(p,point(),center()))
          == SMALLER)
     return ON_BOUNDED_SIDE;
   if (cr == EQUAL)
     return ON_BOUNDARY;
   return ON_UNBOUNDED_SIDE;
  }
  return Bounded_side(compare_T1p(r,squared_distance(center(),p)));
}

template < class R >
inline
bool
Circle_T2<R>::
has_on_boundary(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return bounded_side(p) == ON_BOUNDARY;
}

template < class R >
inline
bool
Circle_T2<R>::
has_on_bounded_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return bounded_side(p) == ON_BOUNDED_SIDE;
}

template < class R >
inline
bool
Circle_T2<R>::
has_on_unbounded_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return bounded_side(p) == ON_UNBOUNDED_SIDE;
}

template < class R >
CGAL_KERNEL_INLINE
bool
Circle_T2<R>::
has_on_negative_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_NEGATIVE_SIDE;
}

template < class R >
CGAL_KERNEL_INLINE
bool
Circle_T2<R>::
has_on_positive_side(const typename Circle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_POSITIVE_SIDE;
}

template < class R >
CGAL_KERNEL_INLINE
std::ostream &
operator<<(std::ostream &os, const Circle_T2<R> &c)
{
  switch(os.iword(IO::mode)) {
  case IO::ASCII :
    os << c.center() << ' ' << c.point() << ' ' <<
       static_cast<int>(c.orientation());
  break;
  case IO::BINARY :
    os << c.center();
    os << c.point();
    write(os, static_cast<int>(c.orientation()));
    break;
  default:
    os << "Circle_T2(" << c.center() <<  ", " << c.point() ;
    switch (c.orientation()) {
      case CLOCKWISE:
        os << ", clockwise)";
        break;
      case COUNTERCLOCKWISE:
        os << ", counterclockwise)";
        break;
      default:
        os << ", collinear)";
        break;
    }
    break;
  }
  return os;
}




template < class R >
CGAL_KERNEL_MEDIUM_INLINE
std::istream &
operator>>(std::istream &is, Circle_T2<R> &c)
{
  
  typename R::Point_T2 center;
  
  /*p will be a point of the boundary of the circle */
  typename R::Point_T2 p;

  int o;

  switch(is.iword(IO::mode)) {
  case IO::ASCII :
    is >> center >> p >> o;
    break;
  case IO::BINARY :
    is >> center;
    is >> p; 
    is >> o;
    break;
  default:
    std::cerr << "" << std::endl;
    std::cerr << "Stream must be in ascii or binary mode" << std::endl;
    break;
  }
  Orientation o1 = Orientation(o);
  if (is) 
    c = Circle_T2<R>(center,p,o1);
  return is;
}

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_CIRCLE_T2_H
