// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Direction_2.h,v $
// $Revision: 1.14 $ $Date: 2003/10/21 12:18:02 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_DIRECTION_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_DIRECTION_T2_H

CGAL_BEGIN_NAMESPACE

template < class R_ >
class Direction_T2
{
  typedef R_                                           R;
  typedef typename R_::FT                              FT;
  typedef typename R_::Vector_T2                       TVector_T2;
  typedef typename R_::Point_T2                        TPoint_T2;
  typedef typename R_::Direction_T2                    TDirection_T2;
  typedef typename R_::Segment_T2                      TSegment_T2;
  typedef typename R_::Line_T2                         TLine_T2;
  typedef typename R_::Aff_transformation_T2           TAff_transformation_T2;
  typedef typename R_::rep                             rep;

  protected:
    Direction_3<rep> *direction;

  public:

    Direction_T2(){}

    //p cannot be antipodal of the origin [1,0,0]
    Direction_T2(const TVector_T2 & p)
    {
      direction = new Direction_3<rep>(p.w(),p.x(),p.y());
    }

    Direction_T2(const TLine_T2 &l)
    {
      TDirection_T2 d = l.direction();
      direction = new Direction_3<rep>(d.dw(),d.dx(),d.dy());
    }

    Direction_T2(const TPoint_T2 &a, const TPoint_T2 &b)

    {
      CGAL_kernel_precondition_msg(!(antipodal(a,b)),
      "'The' direction through antipodal points isn't defined.");
      if (is_at_infinity(a))
				direction = new Direction_3<rep>(a.w(),-a.x(),-a.y());
      else if (is_at_infinity(b))
				direction = new Direction_3<rep>(b.w(),b.x(),b.y());
      else if (is_at_back_range(a,b))
       direction = new Direction_3<rep>(abs(b.w()*a.w()),
			-b.x()*a.w()+a.x()*b.w(),
                              -b.y()*a.w()+a.y()*b.w()) ;
      else
        direction = new Direction_3<rep>(abs(b.w()*a.w()),
			b.x()*a.w()-a.x()*b.w(),
                              b.y()*a.w()-a.y()*b.w()) ;
    }

   //precondition -> s.source() and s.target() != [-1,0,0]
    Direction_T2(const TSegment_T2 &s)
    {
       TPoint_T2 a = s.source();
       TPoint_T2 b = s.target();
       if (is_at_infinity(a))
         direction = new Direction_3<rep>(a.w(),a.x(),a.y());
       else if (is_at_infinity(b))
         direction = new Direction_3<rep>(b.w(),b.x(),b.y());
       else
         direction = new Direction_3<rep>(b.w()*a.w(),b.x()*a.w()-a.x()*b.w(),
                              b.y()*a.w()-a.y()*b.w());
    }

    //the point [w,x,y] cannot be antipodal of the origin [1,0,0]
    Direction_T2(const FT &w, const FT &x, const FT &y)
    {
      direction = new Direction_3<rep>(w,x,y);
    }

    bool    operator==(const TDirection_T2 &d) const;
    bool    operator!=(const TDirection_T2 &d) const;

    TVector_T2       to_vector() const;
    TVector_T2       vector() const { return to_vector(); }


    TDirection_T2  operator-() const;

    const FT & delta(int i) const;

    const FT & dx() const
    {
      return direction->dy();
    }

    const FT & dy() const
    {
      return direction->dz();
    }

    const FT & dw() const
    {
      return direction->dx();
    }

    TDirection_T2 transform(const TAff_transformation_T2 &t) const
    {
      return t.transform(*this);
    }

};

template < class R >
typename Direction_T2<R>::TVector_T2
Direction_T2<R>::to_vector() const
{
  return TVector_T2(dw(),dx(),dy());
}


template < class R >
bool  Direction_T2<R>::operator==(const TDirection_T2 &d) const
{
  return TPoint_T2(FT(0),dx(),dy()) == TPoint_T2(FT(0),d.dx(),d.dy()); //||
}

template < class R >
bool  Direction_T2<R>::operator!=(const TDirection_T2 &d) const
{
  return !(*this==d);
}

template < class R >
typename Direction_T2<R>::TDirection_T2
Direction_T2<R>::operator-() const
{
  return TDirection_T2(dw(),-dx(),-dy());
}


template < class R >
const typename Direction_T2<R>::FT &
Direction_T2<R>::delta(int i) const
{
  CGAL_kernel_precondition( i >= 0 && i <= 2 );
  if (i==0) return dw();
  if (i==1) return dx();
  return dy();
}

//#ifndef CGAL_NO_OSTREAM_INSERT_DIRECTIONC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Direction_T2<R> &d)
{
  switch(os.iword(IO::mode)) {
    case IO::ASCII :
      return os << d.dw() << ' ' << d.dx()  << ' ' << d.dy();
    case IO::BINARY :
      write(os, d.dw());
      write(os, d.dx());
      write(os, d.dy());
      return os;
    default:
      os << "Direction_T2(" << d.dw() << ", " << d.dx() << ", " << d.dy() << ")";
      return os;
  }
}
//#endif // CGAL_NO_OSTREAM_INSERT_DIRECTIONC3

//#ifndef CGAL_NO_ISTREAM_EXTRACT_DIRECTIONC3
template < class R >
std::istream &
operator>>(std::istream &is, Direction_T2<R> &d)
{
  typename R::FT w, x, y;
  switch(is.iword(IO::mode)) {
    case IO::ASCII :
      is >> w >> x >> y;
      break;
    case IO::BINARY :
      read(is, w);
      read(is, x);
      read(is, y);
      break;
    default:
      std::cerr << "" << std::endl;
      std::cerr << "Stream must be in ascii or binary mode" << std::endl;
      break;
  }
  if (is)
      d = Direction_T2<R>(w, x, y);
  return is;
}
//#endif // CGAL_NO_ISTREAM_EXTRACT_DIRECTIONC3


CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_DIRECTION_T2_H
