//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_ISO_RECTANGLE_RANGESEARCH_T2_H
#define CGAL_ISO_RECTANGLE_RANGESEARCH_T2_H

#include <vector>

CGAL_BEGIN_NAMESPACE

template <class Traits, class Point_Iterator>
class Iso_rectangle_RangeSearch_T2 {
  public:
    typedef typename Traits::Point_T2                       Point;
    typedef typename Traits::Point_T1                       Point_T1;
    typedef typename Traits::Line_T2                        Line;
    typedef typename Point::FT                              FT;
    typedef typename Traits::Iso_rectangle_T2               Iso_rectangle;
    typedef std::vector< std::vector <Point> >
                                                            RangeSearch_data_structure;
    typedef typename std::vector<Point>::iterator           vector_point_iterator;
    typedef typename Traits::Front_range_T2                 Front_range_T2;
    typedef typename Traits::Back_range_T2                  Back_range_T2;
    typedef typename Traits::Smaller_y_one_range_infinity_T2
                                                   Smaller_y_one_range_infinity_T2;
    typedef typename Traits::Smaller_x_one_range_infinity_T2
                                                   Smaller_x_one_range_infinity_T2;
    typedef typename Traits::Equal_x_T2                     Equal_x_T2;
    typedef typename Traits::Equal_y_T2                     Equal_y_T2;
    typedef typename Traits::Antipode_T2                    Antipode_T2;
    typedef typename Traits::Antipode_T1                    Antipode_T1;
    typedef typename Traits::Smaller_one_range_infinity_T1
                                                   Smaller_one_range_infinity_T1;
    typedef typename Traits::Equal_T1                       Equal_T1;

  private:
    RangeSearch_data_structure v_rect_dominant_vertex_fr,
                               v_rect_dominant_vertex_br;
    std::vector< std::vector <int> >      v_number_points_inside_rect_fr,
                                          v_number_points_inside_rect_br;
    std::vector< std::vector <bool> >  v_is_input_point_fr,
                                          v_is_input_point_br;

  public:

    Iso_rectangle_RangeSearch_T2(const Point_Iterator begin,
                                 const Point_Iterator end)     {
      RangeSearch_data_structure v_rect_dominant_vertex;
      Point_Iterator it = begin;
      std::vector<Point> front_range_points, back_range_points;
      Front_range_T2 is_at_front_range = Traits().front_range_T2_object();
      Back_range_T2 is_at_back_range = Traits().back_range_T2_object();
      Antipode_T2 antipode = Traits().antipode_T2_object();
      for (; it != end; it++) {
        if (is_at_front_range(*it)) {
          //insert in the structure that holds the points in the front range of T?
          front_range_points.push_back(*it);
        }
        else if (is_at_back_range(*it)) {
          //insert in the structure that holds the points in the back range of T?
          back_range_points.push_back(antipode(*it));
        }
      }
      do_preprocessing(front_range_points.begin(),front_range_points.end(),
                       v_rect_dominant_vertex_fr,
                       v_number_points_inside_rect_fr,
                       v_is_input_point_fr );
      do_preprocessing(back_range_points.begin(),back_range_points.end(),
                       v_rect_dominant_vertex_br,v_number_points_inside_rect_br,
                       v_is_input_point_br );
    }

    int RangeSearch(const Iso_rectangle & r) const
    {
      //if the rectangle is at the back range of T2
      Antipode_T2 antipode_T2 = Traits().antipode_T2_object();
      Antipode_T1 antipode_T1 = Traits().antipode_T1_object();
      if (r.vertex0().w() < 0) {
        Point q(antipode_T2(r.vertex0()));
        Point_T1 n_delta_x(r.x().w(),r.x().x());
        Point_T1 n_delta_y(r.y().w(),r.y().x());
        Iso_rectangle r0 = Iso_rectangle(q,n_delta_x,n_delta_y);
        if (v_rect_dominant_vertex_br.begin() == v_rect_dominant_vertex_br.end())
          return 0;
        return RangeSearch_aux(r0,v_rect_dominant_vertex_br,
                               false,
                               v_is_input_point_br);
      }
      else { //if the rectangle is at the front range of T2
        if (v_rect_dominant_vertex_fr.begin() == v_rect_dominant_vertex_fr.end())
          return 0;
        return RangeSearch_aux(r,v_rect_dominant_vertex_fr,
                              true,
                              v_is_input_point_fr);
      }
    }

  private:

    void do_preprocessing(const vector_point_iterator begin ,
                          const vector_point_iterator end,
                          std::vector< std::vector <Point> >
                                        &  v_rect_dominant_vertex,
                          std::vector< std::vector <int> >
                                        &  v_number_points_inside_rect,
                          std::vector< std::vector <bool> > & v_is_input_point  )
    {
      if (begin == end) //if there is no point from input than return
        return;

      Smaller_x_one_range_infinity_T2 less_x =
                      Traits().smaller_x_one_range_infinity_T2_object();

      int i = 0;
      int j;

      //stores the points sorted in x coordinate
      std::vector< Point > vpts_sorted_in_x;

      //stores the points sorted in y coordinate
      std::vector< Point > vpts_sorted_in_y;

      std::vector< int > v_number_points_line;

      std::vector< Point > v_rect_line;

      vector_point_iterator iterator_x;
      vector_point_iterator iterator_y, previous_iterator_y;
      typename std::vector< std::vector <Point> >::iterator ite;
      typename std::vector< std::vector <int> >::iterator it;
      typename std::vector<int>::iterator it5;
      Smaller_y_one_range_infinity_T2 smaller_y_one_range =
                               Traits().smaller_y_one_range_infinity_T2_object();
      Smaller_x_one_range_infinity_T2 smaller_x_one_range =
                               Traits().smaller_x_one_range_infinity_T2_object();

      std::copy(begin, end, std::back_inserter(vpts_sorted_in_x));
      std::sort(vpts_sorted_in_x.begin(),vpts_sorted_in_x.end(),
                                         smaller_x_one_range);
      std::copy(begin, end, std::back_inserter(vpts_sorted_in_y));
      std::sort(vpts_sorted_in_y.begin(),vpts_sorted_in_y.end(),
                                         smaller_y_one_range);

      v_number_points_line.erase(v_number_points_line.begin(),
                                 v_number_points_line.end());

      Point previous_y(0,0,-1);
      Point previous_x;

      for (iterator_y = vpts_sorted_in_y.begin();
           iterator_y != vpts_sorted_in_y.end();
           iterator_y++)
      {

        if (Point_T1(previous_y.w(),previous_y.y()) !=
            Point_T1(iterator_y->w(),iterator_y->y())) {
          previous_x = Point(0,-1,0);
          v_number_points_line.push_back(0);
          for (iterator_x = vpts_sorted_in_x.begin();
               iterator_x != vpts_sorted_in_x.end();
               iterator_x++) {
            if (Point_T1(previous_x.w(),previous_x.x()) !=
                Point_T1(iterator_x->w(),iterator_x->x())) {
              v_rect_line.push_back(Point(
                         sign(iterator_x->w())*iterator_x->w()*iterator_y->w(),
                         abs(iterator_y->w())*iterator_x->x(),
                         abs(iterator_x->w())*iterator_y->y()));
              v_number_points_line.push_back(0);
              previous_x = *iterator_x;
            }

          }
          previous_y = *iterator_y;
          v_rect_dominant_vertex.push_back(v_rect_line);
          v_rect_line.erase(v_rect_line.begin(),v_rect_line.end());
          v_number_points_inside_rect.push_back(v_number_points_line);
        //  it_aux++;
          v_number_points_line.erase(v_number_points_line.begin(),
                                     v_number_points_line.end());

        }

      }
      unsigned int auxiliar = (*v_number_points_inside_rect.begin()).size();
      for (unsigned int k=0; k < auxiliar;k++)
        v_number_points_line.push_back(0);
      v_number_points_inside_rect.push_back(v_number_points_line);
      v_number_points_line.erase(v_number_points_line.begin(),
                                 v_number_points_line.end());

      ite = v_rect_dominant_vertex.begin();
      std::vector<bool> v_boolean;
      for (; ite != v_rect_dominant_vertex.end(); ite++) {

        for (iterator_x = (*ite).begin(); iterator_x != (*ite).end(); iterator_x++)
          v_boolean.push_back(false);
        v_is_input_point.push_back(v_boolean);
        v_boolean.erase(v_boolean.begin(),v_boolean.end());
      }
      unsigned int index_it = 1;
      int index_prev = 0;
      Point temp;
      for ( ; index_it < vpts_sorted_in_y.size(); index_it++)
      {
        if (vpts_sorted_in_y[index_prev] == vpts_sorted_in_y[index_it]) {
          if (less_x(vpts_sorted_in_y[index_prev],
                     vpts_sorted_in_y[index_it])) {
            temp = *previous_iterator_y;
            vpts_sorted_in_y[index_prev]=*iterator_y;
            vpts_sorted_in_y[index_it]=temp;
          }
        }
        index_prev = index_it;
      }
      iterator_y = vpts_sorted_in_y.begin();
      int k = 0;
      ite = v_rect_dominant_vertex.begin();
      while (iterator_y != vpts_sorted_in_y.end()) {
        bool quit = false;
        if (ite != v_rect_dominant_vertex.end()) {
          iterator_x = (*ite).begin();
          j = 0;

          while (iterator_x != (*ite).end() && !quit) {
            if (*iterator_y == *iterator_x) {
              v_is_input_point[k][j] = true;
              v_number_points_inside_rect[k+1][j+1] = 1;
              quit = true;
            }
            iterator_x++;
            j++;
          }
        }
        previous_iterator_y = iterator_y;
        iterator_y++;
        if (iterator_y != vpts_sorted_in_y.end()) {
          if (!equal_y(*previous_iterator_y,*iterator_y)) {
            k++;
            if (ite != v_rect_dominant_vertex.end()) {
              ite++;
            }
          }
        }
      }
      it = v_number_points_inside_rect.begin();
      it++;
      i = 1;
      for (; it!=v_number_points_inside_rect.end(); it++) {
        j = 1;
        it5 = (*it).begin();
        it5++;
        for (; it5 != (*it).end(); it5++) {
          v_number_points_inside_rect[i][j] +=
                          v_number_points_inside_rect[i][j-1] +
                          v_number_points_inside_rect[i-1][j] -
                          v_number_points_inside_rect[i-1][j-1];
          j++;
        }
        i++;
      }
    }



    int RangeSearch_aux(const Iso_rectangle r,
                        const std::vector< std::vector <Point> >
                                          v_rect_dominant_vertex,
                        const bool front_range,
                        std::vector< std::vector <bool> > v_is_input_point) const
    {
      std::vector< std::vector <int> >
                 v_number_points_inside_rect;
      if (front_range)
        v_number_points_inside_rect = v_number_points_inside_rect_fr;
      else
        v_number_points_inside_rect = v_number_points_inside_rect_br;

      Smaller_x_one_range_infinity_T2 less_x =
                      Traits().smaller_x_one_range_infinity_T2_object();

      Smaller_y_one_range_infinity_T2 less_y =
                      Traits().smaller_y_one_range_infinity_T2_object();

      Point smaller_coordinate_x,larger_coordinate_x;

      Point smaller_coordinate_y,larger_coordinate_y;

      std::vector< Point > vector_y;

      if (less_x(r.vertex0(),r.vertex1())) {
        smaller_coordinate_x = r.vertex0();
        larger_coordinate_x = r.vertex1();
      }
      else {
        larger_coordinate_x = r.vertex0();
        smaller_coordinate_x = r.vertex1();
      }
      if (less_y(r.vertex0(),r.vertex3())) {
        smaller_coordinate_y = r.vertex0();
        larger_coordinate_y = r.vertex3();
      }
      else {
        larger_coordinate_y = r.vertex0();
        smaller_coordinate_y = r.vertex3();
      }
      int q1 = -1;
      int q2 = -1;
      int q3 = -1;
      int q4 = -1;
      int index_smaller_x = -1;
      int index_smaller_y = -1;
      int index_larger_x = -1;
      int index_larger_y = -1;
      std::vector<Point> v_aux = v_rect_dominant_vertex[0];
      typename std::vector< std::vector<Point> >::const_iterator it;
      typename std::vector<Point>::const_iterator it2;
      for (it = v_rect_dominant_vertex.begin();
           it != v_rect_dominant_vertex.end(); it++)
      {
        it2 = (*it).begin();
        vector_y.push_back(*it2);
        it2++;
      }
      if (r.x() == Point_T1(0,-1)) {
        q2 = 0;
        q3 = 0;
      }
      if (r.y() == Point_T1(0,-1)) {
        q3 = 0;
        q4 = 0;
      }
      if (less_x(smaller_coordinate_x,v_aux[0])) {
        q2 = 0;
        q3 = 0;
      }
      if (less_y(smaller_coordinate_y,vector_y[0])) {
        q3 = 0;
        q4 = 0;
      }
      if (r.x() == Point_T1(0,1)) {
        typename std::vector<std::vector <int> >::iterator
               it0 = v_number_points_inside_rect.begin();
        std::vector<int> v;
        v = (*it0);
        index_larger_x =  v.size()-1;
      }
      if (r.y() == Point_T1(0,1)) {
        index_larger_y =  v_number_points_inside_rect.size()-1;
      }
      if (q2 == 0 && q3 == 0)
        index_smaller_x = 0;
      if (q3 == 0 && q4 == 0)
        index_smaller_y = 0;
      if (index_smaller_x == -1) {
        index_smaller_x = Binary_search(0,v_aux.size()-1,
                                        v_aux,
                                        Point_T1(smaller_coordinate_x.w(),
                                                 smaller_coordinate_x.x()),
                                        1);
        if (index_smaller_x > 0) {
          if (less_x(smaller_coordinate_x,v_aux[index_smaller_x]))
            index_smaller_x--;
        }
      }
      else {
        index_smaller_x = 0;
      }
      if (index_larger_x == -1) {
        index_larger_x = Binary_search(0,v_aux.size()-1,
                                       v_aux,Point_T1(larger_coordinate_x.w(),
                                       larger_coordinate_x.x()),
                                       1);
        int v_rect_size = v_rect_dominant_vertex[0].size();
        if (index_larger_x < v_rect_size) {
          if (less_x(v_aux[index_larger_x],larger_coordinate_x))
            index_larger_x++;
        }
      }
      else {
        index_larger_x = v_rect_dominant_vertex[0].size();
      }
      if (index_smaller_y == -1) {
        index_smaller_y = Binary_search(0,vector_y.size()-1,
                                        vector_y,
                                        Point_T1(smaller_coordinate_y.w(),
                                                 smaller_coordinate_y.y()),
                                        2);

        if (index_smaller_y > 0) {
          if (less_y(smaller_coordinate_y,vector_y[index_smaller_y]))
            index_smaller_y--;
        }
      }
      else {
        index_smaller_y = 0;
      }
      if (index_larger_y == -1) {
        index_larger_y = Binary_search(0,vector_y.size()-1,
                                       vector_y,
                                       Point_T1(larger_coordinate_y.w(),
                                                larger_coordinate_y.y()),
                                       2);
        int vector_y_size = vector_y.size();
        if (index_larger_y < vector_y_size) {
          if (less_y(vector_y[index_larger_y],larger_coordinate_y))
            index_larger_y++;
        }
      }
      else {
        index_larger_y = vector_y.size();
      }
      if (q3 == -1){
        q3 = v_number_points_inside_rect[index_smaller_y+1][index_smaller_x+1];
      }
      if (q1 == -1) {
        q1 = v_number_points_inside_rect[index_larger_y][index_larger_x];
      }
      if (q2 == -1) {
        q2 = v_number_points_inside_rect[index_larger_y][index_smaller_x+1];
      }
      if (q4 == -1) {
        q4 = v_number_points_inside_rect[index_smaller_y+1][index_larger_x];
      }
      return ((q1-q2-q4+q3) < 0 ? 0 : (q1-q2-q4+q3));
    }



    int Binary_search(int left, int right, std::vector<Point> & v,
                      const Point_T1 p, int index) const
    {
      Smaller_one_range_infinity_T1 less =
                       Traits().smaller_one_range_infinity_T1_object();
      Equal_T1  equal = Traits().equal_T1_object();
      if ((right - left) <= 0) {
        return left;
      }
      else if (equal(Point_T1(v[(left + right)/2].w(),v[(left + right)/2][index]),p))
        return (left + right)/2;
      else if (less(p,Point_T1(v[(left + right)/2].w(),v[(left + right)/2][index]))) {
        return Binary_search(left,(left + right)/2 - 1,v,p,index);
      }
      else {
        return Binary_search((left + right)/2 + 1,right,v,p,index);
      }
    }
};

CGAL_END_NAMESPACE

#endif
