//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_LARGEST_EMPTY_CIRCLE_T2_H
#define CGAL_LARGEST_EMPTY_CIRCLE_T2_H


#include <iostream>

#include <CGAL/Delaunay_triangulation_T2.h>

CGAL_BEGIN_NAMESPACE

template < class Gt, 
           class Tds = Triangulation_data_structure_T2 <
                             Triangulation_vertex_base_2<Gt>,
                             Triangulation_face_base_2<Gt, Triangulation_ds_face_base_T2<> > > >
class Largest_empty_circle_T2
  : private Delaunay_triangulation_T2<Gt,Tds>
{
  typedef Delaunay_triangulation_T2<Gt,Tds>  Delaunay_triangulation;
  typedef Tds                                Triangulation_data_structure;
  typedef Gt                                 Geom_traits;
  typedef typename Gt::Circle_T2	     Circle;
  typedef typename Delaunay_triangulation::Finite_faces_iterator Finite_faces_iterator;
  typedef typename Delaunay_triangulation::Face_handle Face_handle;
  typedef typename Delaunay_triangulation::Face_circulator Face_circulator;
  typedef typename Delaunay_triangulation::Point Point;

  Geom_traits gt;
  Circle largest_empty_circle;
  
  public:

  Largest_empty_circle_T2(const Geom_traits& geom_traits=Geom_traits()) {
    gt = geom_traits;
  }

  Circle get_circle() const
  {
    return largest_empty_circle;
  }
  
  template <class InputIterator>
  bool compute(InputIterator first, InputIterator last) 
  {
    int i;
    insert(first,last);
    Finite_faces_iterator fi = finite_faces_begin();
       
    Circle candidate_circle;
    bool found = false;
    while (!found && fi != finite_faces_end()) {
      if (!make_circle(fi,largest_empty_circle)) {
        fi++;
      }
      else {
        found = true;
      }
    }
    if (!found) return false;
    for (; fi != finite_faces_end(); fi++) {
      if (make_circle(fi,candidate_circle)) {
        if (gt.compare_radius(candidate_circle,
                              largest_empty_circle) == LARGER) {
          largest_empty_circle = candidate_circle;
        }
      }
    }

    if (ch_is_defined) {
      Face_circulator fc = incident_faces(infinite_vertex()), done(fc);
      do {
        i = fc->index(infinite_vertex());
        if (make_circle(fc->vertex(cw(i))->point(),
                        fc->vertex(ccw(i))->point(),candidate_circle)) {
          if (gt.compare_radius(candidate_circle,
			     largest_empty_circle) == LARGER) {
	    largest_empty_circle = candidate_circle;
          }
	}
      } while(fc!=done);	 
    }
    return true;
  }
  
  bool make_circle (const Point &p, const Point &q, Circle &cir) const
  {
    Point c;
    if (gt.exist_circle(p,q,c)) {	
      if (sign(c.w()) != sign(p.w()))       
        c = Point(-c.w(),-c.x(),-c.y());
      cir = Circle(c,p);
      return true;
    }	
    return false;
  }
  

  bool make_circle (Face_handle fh, Circle &cir) const
  {
    Point c;
    if (is_infinite(fh)) {
      int index = fh->index(infinite_vertex());
      if (gt.exist_circle(fh->vertex(cw(index))->point(),
                          fh->vertex((ccw(index)))->point(),c)) {
        if (sign(c.w()) != sign(fh->vertex(cw(index))->point().w()))       
          c = Point(-c.w(),-c.x(),-c.y());
        cir = Circle(c,fh->vertex(cw(index))->point());
        return true;
      }
      return false;
    }	
    else {
      if (gt.exist_circle(fh->vertex(0)->point(),
    	        	      fh->vertex(1)->point(),
			      fh->vertex(2)->point(),c)) {
        if (sign(c.w()) != sign(fh->vertex(0)->point().w()))       
          c = Point(-c.w(),-c.x(),-c.y()); 
        cir = Circle (c,fh->vertex(0)->point());
        return true;
     }
     return false;
   }
  }
  
};

template < class Gt, class Tds > std::istream& operator>>
    (std::istream& is, Largest_empty_circle_T2<Gt,Tds> &lec);
template < class Gt, class Tds >  std::ostream& operator<<
  (std::ostream& os, const Largest_empty_circle_T2<Gt,Tds> &lec);


template <class Gt, class Tds >
std::ostream&
operator<<(std::ostream& os, const Largest_empty_circle_T2<Gt,Tds> &lec)
{
  typedef typename Gt::Circle_T2 Circle;
  Circle c = lec.get_circle();
  switch(os.iword(IO::mode)) {
  case IO::ASCII :
    os << c.center() << ' ' << c.squared_radius() << ' ' <<
       static_cast<int>(c.orientation());
  break;
  case IO::BINARY :
    os << c.center();
    os << c.squared_radius();
    write(os, static_cast<int>(c.orientation()));
    break;
  default:
    os << "Circle_T2(" << c.center() <<  ", " << c.squared_radius() ;
    switch (c.orientation()) {
      case CLOCKWISE:
        os << ", clockwise)";
        break;
      case COUNTERCLOCKWISE:
        os << ", counterclockwise)";
        break;
      default:
        os << ", collinear)";
        break;
    }
    break;
  }
  return os;
}



template < class Gt, class Tds >
std::istream&
operator>>(std::istream& is, Largest_empty_circle_T2<Gt,Tds> &lec)
{
  typedef typename Gt::Circle_T2 Circle;

  typename Gt::Point_T2 center;
  typename Gt::Point_T1 squared_radius;
  
  int o;
  switch(is.iword(IO::mode)) {
  case IO::ASCII :
    is >> center >> squared_radius >> o;
    break;
  case IO::BINARY :
    is >> center;
    is >> squared_radius; 
    is >> o;
    break;
  default:
    std::cerr << "" << std::endl;
    std::cerr << "Stream must be in ascii or binary mode" << std::endl;
    break;
  }
	Orientation o1 = Orientation(o);
  if (is) 
    lec = Circle(center, squared_radius,o1);
  return is;
}

  
CGAL_END_NAMESPACE  
  
#endif
  
