// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Line_2.h,v $
// $Revision: 1.16 $ $Date: 2003/10/21 12:18:07 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_LINE_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_LINE_T2_H

CGAL_BEGIN_NAMESPACE

template < class R_ >
class Line_T2 : Handle_for<std::pair< typename R_::Point_T2,
																			typename R_::Point_T2> >
{
  typedef typename R_::FT                       FT;
  typedef typename R_::RT                       RT;
  typedef typename R_::Point_T2                 TPoint_T2;
  typedef typename R_::Point_T1                 TPoint_T1;
  typedef typename R_::Vector_T2                TVector_T2;
  typedef typename R_::Direction_T2             TDirection_T2;
  typedef typename R_::Ray_T2                   TRay_T2;
  typedef typename R_::Line_T2                  TLine_T2;
  typedef typename R_::Segment_T2               TSegment_T2;
  typedef typename R_::Aff_transformation_T2    TAff_transformation_T2;
  typedef typename R_::Line_handle_T2           base;
  typedef typename base::element_type           rep;


public:
  typedef R_                                     R;

  Line_T2(){}

  Line_T2(const TPoint_T2 &p, const TPoint_T2 &q)
  {
    CGAL_kernel_precondition_msg(!antipodal(p,q),
                                 "The 'line' conecting p and q isn't defined");
    CGAL_kernel_precondition_msg(p!=q,
              "A 'line' is determined by two diferent and not antipodal points");
    if (p.w() == 0)
      initialize_with(rep(q,antipode(p)));
    else
      initialize_with(rep(p,q));
  }

  Line_T2(const TSegment_T2 &s)
  {
    TPoint_T2 s0 = s.start();
    TPoint_T2 s1 = s.target();
    CGAL_kernel_precondition_msg(s0!=s1,
              "A 'line' is determined by two diferent and not antipodal points");
    initialize_with(rep(s0,s1));
  }

  Line_T2(const TRay_T2 &r)
  {
    TPoint_T2 p = r.start();
    TPoint_T2 q = r.second_point();
    initialize_with(rep(p,q));
  }

  Line_T2(const TPoint_T2 &p, const TDirection_T2 &d)
  {
    CGAL_kernel_precondition(!is_at_infinity(p));
    TPoint_T2 *q = new TPoint_T2(FT(0),d.dx(),d.dy());
    CGAL_kernel_precondition_msg(!antipodal(p,*q),
    "The 'line' that contains p and goes in its antipodal direction isn't defined");
    initialize_with(rep(p,*q));   
  }

  bool        operator==(const Line_T2 &l) const;
  bool        operator!=(const Line_T2 &l) const;

  const FT W() const;

  const FT X() const;

  const FT Y() const;

  TLine_T2 opposite() const;

  const TPoint_T2 &  point0() const
  {
    return Ptr()->first;
  }

  const TPoint_T2 &  point1() const
  {
    return Ptr()->second;
  }

  const TDirection_T2 & direction() const
  {
    CGAL_kernel_precondition(!is_at_infinity(point0()) ||
                             !is_at_infinity(point1()));
    TDirection_T2 *d;
    d = new TDirection_T2(point1()-point0());
    return *d;
  }

  TPoint_T2     point(int i) const;

  TPoint_T2     projection(const TPoint_T2 &p) const;

  TVector_T2    to_vector() const;

  bool has_on(const TPoint_T2 &p) const;

  bool has_on_positive_side(const TPoint_T2 &p) const;

  bool has_on_negative_side(const TPoint_T2 &p) const;

  Oriented_side oriented_side(const TPoint_T2 &p) const;

  bool has_on_boundary(const TPoint_T2 &p) const;

  bool is_degenerate() const;

  TLine_T2 transform(const TAff_transformation_T2 &t) const
  {
    return TLine_T2(t.transform(point0()),t.transform(point1()));
  }
  
  const FT x_at_wy(const FT &w , const FT &y) const;

  const FT y_at_wx(const FT &w , const FT &x) const;

  const FT w_at_xy(const FT &x , const FT &y) const;

};

template < class R >
inline
bool
Line_T2<R>::operator==(const Line_T2<R> &l) const
{
  return (dual(l) == dual(*this));
}

template < class R >
inline
bool
Line_T2<R>::operator!=(const Line_T2<R> &l) const
{
  return !(*this == l);
}


template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::W() const
{
  return FT(this->point0().x()*this->point1().y()-
            this->point0().y()*this->point1().x());
}

template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::X() const
{
  return FT(-(this->point0().w()*this->point1().y()
             -this->point0().y()*this->point1().w()));
}

template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::Y() const
{
  return FT(this->point0().w()*this->point1().x()
          -this->point0().x()*this->point1().w());
}

template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::x_at_wy(
                   const typename Line_T2<R>::FT &w,
                   const typename Line_T2<R>::FT &y) const
{
  CGAL_kernel_precondition(X()!=0);
  return ((-W()*w-Y()*y)/X());
}

template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::y_at_wx(
                   const typename Line_T2<R>::FT &w,
                   const typename Line_T2<R>::FT &x) const
{
  CGAL_kernel_precondition(Y()!=0);
  return ((-W()*w-X()*x)/Y());
}

template < class R >
inline
const typename Line_T2<R>::FT Line_T2<R>::w_at_xy(
                   const typename Line_T2<R>::FT &x,
                   const typename Line_T2<R>::FT &y) const
{
  CGAL_kernel_precondition(W()!=0);
  return ((-X()*x-Y()*y)/W());
}

template < class R >
inline
typename Line_T2<R>::TPoint_T2
Line_T2<R>::
point(int i) const
{
  return point_on_line(i, *this);
}


template < class R >
inline
typename Line_T2<R>::TLine_T2
Line_T2<R>::opposite() const
{
  return Line_T2<R>(point1(), point0());
}


template < class R >
inline
typename Line_T2<R>::TPoint_T2
Line_T2<R>::
projection(const typename Line_T2<R>::TPoint_T2 &p) const
{
  return projection_line(p, *this);
}

template < class R >
inline
bool
Line_T2<R>::
has_on(const typename Line_T2<R>::TPoint_T2 &p) const
{
  TPoint_T2 a = point0();
  TPoint_T2 b = point1();
  return collinear(a, b, p);
}

template < class R >
CGAL_KERNEL_INLINE
Oriented_side
Line_T2<R>::
oriented_side(const typename Line_T2<R>::TPoint_T2 &p) const
{
  return side_of_oriented_line(*this, p);
}

template < class R >
inline
bool
Line_T2<R>::
has_on_positive_side(const typename Line_T2<R>::TPoint_T2 &p) const
{
  //std::cout << "has_on_positive_side \n";
  return oriented_side(p) == ON_POSITIVE_SIDE;
}

template < class R >
inline
bool
Line_T2<R>::
has_on_negative_side(const typename Line_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_NEGATIVE_SIDE;
}

template < class R >
inline
bool
Line_T2<R>::
has_on_boundary(const typename Line_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_ORIENTED_BOUNDARY;
}

template < class R >
inline
bool
Line_T2<R>::is_degenerate() const
{ 
  return point0() == point1();
}

template < class R >
inline
typename Line_T2<R>::TVector_T2
Line_T2<R>::to_vector() const
{ 
  return TVector_T2(direction());
}

//#ifndef CGAL_CARTESIAN_NO_OSTREAM_INSERT_LINEC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Line_T2<R> &l)
{
  switch(os.iword(IO::mode)) {
  case IO::ASCII :
    return os << l.point0() << ' ' << l.point1();
  case IO::BINARY :
    return os << l.point0() <<  l.point1();
  default:
    return  os << "Line_T2(" << l.point0() << ", " << l.point1() << ")";
  }
}
//#endif // CGAL_CARTESIAN_NO_OSTREAM_INSERT_LINEC3

#ifndef CGAL_CARTESIAN_NO_ISTREAM_EXTRACT_LINEC3
template < class R >
std::istream &
operator>>(std::istream &is, Line_T2<R> &l)
{
  typename R::Point_T2 p, q;
  is >> p >> q;
  if (is) 
    l = Line_T2<R>(p, q);
  return is;
}
#endif // CGAL_CARTESIAN_NO_ISTREAM_EXTRACT_LINEC3

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_LINE_T2_H
