// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Planar_map/include/CGAL/Pm_segment_traits_2.h,v $
// $Revision: 1.31 $ $Date: 2003/09/18 10:24:35 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Iddo Hanniel      <hanniel@math.tau.ac.il>
//                 Eyal Flato        <flato@post.tau.ac.il>
//                 Oren Nechushtan   <theoren@math.tau.ac.il>
//                 Eti Ezra          <estere@post.tau.ac.il>
//                 Shai Hirsch       <shaihi@post.tau.ac.il>
//                 Eugene Lipovetsky <eug@post.tau.ac.il>
//                 Efi Fogel         <efif@post.tau.ac.il>
//                 Ron Wein          <wein@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PM_SEGMENT_TRAITS_T2_H
#define CGAL_PM_SEGMENT_TRAITS_T2_H

#include <CGAL/tags.h>

CGAL_BEGIN_NAMESPACE

template <class Kernel_>
class Pm_segment_traits_T2 : public Kernel_
{
public:
  typedef Kernel_                         Kernel;

#if !defined(HAS_LEFT_NOT)
  typedef Tag_true                        Has_left_category;
#else
  typedef Tag_false                       Has_left_category;
#endif
    
  // Traits objects
  typedef typename Kernel::Point_T2        Point_T2;
  typedef typename Kernel::Segment_T2      X_monotone_curve_T2;

  // Backward compatability    
  typedef Point_T2                         Point;
  typedef X_monotone_curve_T2              X_curve;

protected:
  // Functors:
  typedef typename Kernel::Construct_vertex_T2               Construct_vertex_T2;
  typedef typename Kernel::Equal_T2                          Equal_T2;
  typedef typename Kernel::Intersect_T2                      Intersect_T2;
  typedef typename Kernel::Cmp_dist_collinear_origin_T2      Cmp_dist_collinear_origin_T2;
  typedef typename Kernel::Line_T2                           Line_T2;
  typedef typename Kernel::Object_T2                         Object_T2;
  typedef typename Kernel::Has_on_segment_T2                 Has_on_segment_T2;

    
public:
  // Creation
  Pm_segment_traits_T2() {}

  // Operations
  // ----------
    
  /*! Get the curve source.
   * \param cv The curve.
   * \return The source point.
   */
  Point_T2 curve_source(const X_monotone_curve_T2 & cv) const 
  { return construct_vertex_T2_object()(cv, 0); }

  /*! Get the curve target.
   * \param cv The curve.
   * \return The target point.
   */
  Point_T2 curve_target(const X_monotone_curve_T2 & cv) const 
  { return construct_vertex_T2_object()(cv, 1); }


 /*! Check if the two points are the same.
   * \param p1 The first point.
   * \param p2 The second point.
   * \return (true) if p1 == p2.
   */

  bool point_equal(const Point_T2 & p1, const Point_T2 & p2) const
  { return equal_T2_object()(p1, p2); }

    
/*! curve_overlapSweepLine()
   * \param cv the curve
   * \return true iff the curve overlaps sweep line
   */
  bool curve_overlapSweepLine(const X_monotone_curve_T2 & cv) const 
  { 
    typedef typename Point_T2::FT FT;

    Point_T2 o(FT(1),FT(0),FT(0));
    return (orientation(o,cv.source(),cv.target()) == 0); 
  }

  bool curve_overlapSweepLine_at_currentPos(const X_monotone_curve_T2 & cv,
                                            const Point_T2 & pos) const 
  { 
    typedef typename Point_T2::FT FT;

    Point_T2 o(FT(1),FT(0),FT(0));
    
    return (orientation(o,cv.source(),pos) == COLLINEAR && 
            orientation(o,cv.target(),pos) == COLLINEAR); 
  }



  Comparison_result curves_compare_forward(const X_monotone_curve_T2 & cv1,
                                           const X_monotone_curve_T2 & cv2,
                                           const Point_T2 & q) const
  {

   
    Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
    const Point_T2 & source1 = construct_vertex(cv1, 0);
    const Point_T2 & target1 = construct_vertex(cv1, 1);
    const Point_T2 & source2 = construct_vertex(cv2, 0);
    const Point_T2 & target2 = construct_vertex(cv2, 1);   

    typedef typename Point_T2::FT FT;
    const Point_T2 o(FT(1),FT(0),FT(0));
    const Point_T2 ao(FT(-1),FT(0),FT(0));
    Point_T2 second1,second2;
    if ((q == o) || (q == ao)) {
      second1 = (source1 != q) ? source1 : target1;  
      second2 = (source2 != q) ? source2 : target2;
      return Comparison_result(orientation(second1,o,second2));
    }
    if (orientation(o,q,source1) != 1)
      second1 = target1;
    else
      second1 = source1;
    if (orientation(o,q,source2) != 1)
      second2 = target2;
    else
      second2 = source2;
    return Comparison_result(orientation(q,second1,second2));
  }

  Comparison_result curves_compare_backward(
                          const X_monotone_curve_T2 & cv1,
                          const X_monotone_curve_T2 & cv2,
                          const Point_T2 & q) const
  {
    CGAL_precondition(! curve_overlapSweepLine(cv1));
    CGAL_precondition(! curve_overlapSweepLine(cv2));


    Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
    const Point_T2 & source1 = construct_vertex(cv1, 0);
    const Point_T2 & target1 = construct_vertex(cv1, 1);
    const Point_T2 & source2 = construct_vertex(cv2, 0);
    const Point_T2 & target2 = construct_vertex(cv2, 1);   

    typedef typename Point_T2::FT FT;
    const Point_T2 o(FT(1),FT(0),FT(0));
    Point_T2 second1,second2;

    if (orientation(o,q,source1) != -1)
      second1 = target1;
    else
      second1 = source1;
    if (orientation(o,q,source2) != -1)
      second2 = target2;
    else
      second2 = source2;

    return Comparison_result(orientation(q,second1,second2));
  }



 /*! point_in_PerpendicularDirSweep_range()
   * \param cv the curve
   * \param q the point
   * \return true if q is in the x range of cv
   */
  bool point_in_PerpendicularDirSweep_range(
                                    const X_monotone_curve_T2 & cv,
                                    const Point_T2 & q) const
  {
    typedef typename Point_T2::FT FT;
    const Point_T2 o(FT(1),FT(0),FT(0));
    const Point_T2 source = cv.source();
    const Point_T2 target = cv.target();
    Point_T2 firstTobeProcessed;
    Point_T2 secondTobeProcessed;
    if (orientation(o,source,target) == 1) {
      firstTobeProcessed = source;
      secondTobeProcessed = target;
    }
    else if (orientation(o,source,target) == 0) {
      if (source != Point_T2(FT(1),FT(0),FT(0)) && source != Point_T2(FT(-1),FT(0),FT(0))) 
        return (orientation(o,source,q)==0);
      return (orientation(o,target,q)==0);
    }
    else {
      firstTobeProcessed = target;
      secondTobeProcessed = source;
    }
    return ((orientation(o,firstTobeProcessed,q) != -1) &&
            (orientation(o,secondTobeProcessed,q) != 1));
  }


  Comparison_result curves_compare_y_at_x_right(const X_monotone_curve_T2 & cv1,
                                                const X_monotone_curve_T2 & cv2, 
                                                const Point_T2 & q) const 
  {
    // The two curves must not be vertical.
    CGAL_precondition(! curve_overlapSweepLine(cv1));
    CGAL_precondition(! curve_overlapSweepLine(cv2));

    // The two curve must be defined at q and also to its left.
    Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
    const Point_T2 & source1 = construct_vertex(cv1, 0);
    const Point_T2 & target1 = construct_vertex(cv1, 1);
    const Point_T2 & source2 = construct_vertex(cv2, 0);
    const Point_T2 & target2 = construct_vertex(cv2, 1);

    typedef typename Point_T2::FT FT;
    Point_T2 o(FT(1),FT(0),FT(0));
    CGAL_precondition (int(orientation(o,q,source1)) * int(orientation(o,q,target1))!=1);
    CGAL_precondition (int(orientation(o,q,source2)) * int(orientation(o,q,target2))!=1);
    
       
    // Since the curves are continuous, if they are not equal at q, the same
    // result also applies to q's left.
    //CGAL_precondition (compare_y_at_x_2_object()(q, cv1, cv2) == EQUAL);
    
    // <cv2> and <cv1> meet at a point with the same x-coordinate as q
    // compare their derivatives.
    
    Orientation ori1 = orientation(o,q,source1);
    Orientation ori2 = orientation(o,q,target1);

    Point_T2 p1,p2;

    if (ori1 == 1)
      p1 = source1;
    else if (ori2 == 1)
      p1 = target1;
    else {
      if (ori1 == 0)
        p1 = source1;
      else
        p1 = target1;
    }
     
    ori1 = orientation(o,q,source2);
    ori2 = orientation(o,q,target2);

    if (ori1 == 1)
      p2 = source2;
    else if (ori2 == 1)
      p2 = target2;
    else {
      if (ori1 == 0)
        p2 = source2;
      else
        p2 = target2;
    }
    return Comparison_result(orientation(q,p1,p2));
  }


  Comparison_result curves_compare_y_at_x_left(const X_monotone_curve_T2 & cv1,
                                               const X_monotone_curve_T2 & cv2, 
                                               const Point_T2 & q) const 
  {
    // The two curves must not be vertical.
    CGAL_precondition(!curve_overlapSweepLine(cv1));
    CGAL_precondition(!curve_overlapSweepLine(cv2));

    // The two curve must be defined at q and also to its left.
    Construct_vertex_T2 construct_vertex = construct_vertex_T2_object();
    const Point_T2 & source1 = construct_vertex(cv1, 0);
    const Point_T2 & target1 = construct_vertex(cv1, 1);
    const Point_T2 & source2 = construct_vertex(cv2, 0);
    const Point_T2 & target2 = construct_vertex(cv2, 1);

    typedef typename Point_T2::FT FT;
    Point_T2 o(FT(1),FT(0),FT(0));
    CGAL_precondition (int(orientation(o,q,source1)) * int(orientation(o,q,target1))!=1);
    CGAL_precondition (int(orientation(o,q,source2)) * int(orientation(o,q,target2))!=1);
    
       
    // Since the curves are continuous, if they are not equal at q, the same
    // result also applies to q's left.
    //CGAL_precondition (compare_y_at_x_2_object()(q, cv1, cv2) == EQUAL);
    
    // <cv2> and <cv1> meet at a point with the same x-coordinate as q
    // compare their derivatives.
    
    Orientation ori1 = orientation(o,q,source1);
    Orientation ori2 = orientation(o,q,target1);

    Point_T2 p1,p2;

    if (ori1 == -1)
      p1 = source1;
    else if (ori2 == -1)
      p1 = target1;
    else {
      if (ori1 == 0)
        p1 = source1;
      else
        p1 = target1;
    }
     
    ori1 = orientation(o,q,source2);
    ori2 = orientation(o,q,target2);

    if (ori1 == -1)
      p2 = source2;
    else if (ori2 == -1)
      p2 = target2;
    else {
      if (ori1 == 0)
        p2 = source2;
      else
        p2 = target2;
    }
    return Comparison_result(orientation(q,p1,p2));
  }


  /*! curve_compare_dirSweepLine_at_perpendicularDir() compares the
		 y-coordinate of two given curves at
   * the x-coordinate of a given point.
   * \param cv1 the first curve
   * \param cv2 the second curve
   * \param q the point
   * \return LARGER if cv1(x(q)) > cv2(x(q)); SMALLER if cv1(x(q)) < cv2(x(q));
   * or else EQUAL.
   * \pre The point q is in the x range of the two curves.
   */
  Comparison_result curve_compare_dirSweepLine_at_perpendicularDir(
                                            const Point_T2 & q,
                                            const X_monotone_curve_T2 & cv1) const
  {
    CGAL_precondition(point_in_PerpendicularDirSweep_range(cv1, q)); 

    Cmp_dist_collinear_origin_T2 cmp_dist_collinear_origin = cmp_dist_collinear_origin_T2_object();
    Has_on_segment_T2 has_on = has_on_segment_T2_object();

    if (curve_overlapSweepLine(cv1)) {

      if (has_on(cv1,q))
        return EQUAL;

      return cmp_dist_collinear_origin(q,cv1.source());
    }
    typedef typename Point_T2::FT FT;
    Point_T2 o(FT(1),FT(0),FT(0));
    if (q == o) {

      if (has_on(cv1,q))
        return EQUAL;

      return SMALLER;
    }
    if (q == Point_T2(FT(-1),FT(0),FT(0))) {

      return LARGER;
    }
    Intersect_T2 intersect = intersect_T2_object();

    Line_T2 l(o,q);

    Point_T2 p;

    Object_T2 obj = intersection(l,cv1);

    assign(p,obj);
    return cmp_dist_collinear_origin(q,p);
  }

  Comparison_result curves_compare_dirSweepLine_at_perpendicularDir(
                                         const X_monotone_curve_T2 & cv1,
                                         const X_monotone_curve_T2 & cv2,
                                         const Point_T2 & q) const
  {
    CGAL_precondition(point_in_PerpendicularDirSweep_range(cv1, q));
    CGAL_precondition(point_in_PerpendicularDirSweep_range(cv2, q));
    Cmp_dist_collinear_origin_T2 cmp_dist_collinear_origin = 
                                 cmp_dist_collinear_origin_T2_object();
    
    typedef typename Point_T2::FT FT;
    Point_T2 o(FT(1),FT(0),FT(0));
    Intersect_T2 intersect = intersect_T2_object();

    Object_T2 obj;
    Point_T2 p1,p2;
    if (antipode(o) == q) {
      obj = intersection(q,cv1);
      assign(p1,obj);
      obj = intersection(q,cv2);
      assign(p2,obj);  
    }
    else {

      Line_T2 l(o,q);

      obj = intersection(l,cv1);

      assign(p1,obj);
      obj = intersection(l,cv2);

      assign(p2,obj); 
    }
    if (cv1.has_on(q) && cv2.has_on(q))
      return EQUAL;
    return cmp_dist_collinear_origin(p1,p2);
  }

  bool compare_xy(const Point_T2 & p1,const Point_T2 & p2) const
  {
    return compare_around_origin_T2_object()(p1,p2); 
  }

};

CGAL_END_NAMESPACE

#endif
