// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Point_3.h,v $
// $Revision: 1.15 $ $Date: 2003/10/21 12:18:10 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_POINT_T2_H
#define CGAL_POINT_T2_H

#include <CGAL/Point_3.h>

CGAL_BEGIN_NAMESPACE

template < class R_ > class Point_T2 
{
	public:
  typedef R_ R;
  typedef typename R::FT                      FT;
  typedef typename R::rep	 	                  rep;
  typedef typename R::Vector_T2               TVector_T2;
  typedef typename R::Point_T2                TPoint_T2;
  typedef typename R::Aff_transformation_T2   TAff_transformation_T2;

  
  Point_T2(){}

  Point_T2(const FT &w, const FT &x, const FT &y)
  { 
    CGAL_kernel_precondition_msg( ((w!=0) || (x!=0) || (y!=0)),
                                     "Invalid Triple in T^2!");
    Point=new Point_3<rep>(w,x,y);
  }
  
  Point_T2(const TVector_T2 & v)
  {
    CGAL_kernel_precondition_msg( ((v.w()!=0) || (v.x()!=0) || (v.y()!=0)),
                                     "Invalid Triple in T^2!");
    Point=new Point_3<rep>(v.w(),v.x(),v.y());
  }
 
  const FT & w() const
  {
     return Point->x();
  }
  
  const FT & x() const
  {
     return Point->y();
  }

  const FT & y() const
  {
     return Point->z();
  }

  bool operator==(const Point_T2 &p) const
  {
    if ((p.w() != 0) || (this->w()!=0)) {
      if (!((p.w() * this->x()) == (p.x() * this->w())))
           return false;
      if (!((p.w() * this->y()) == (p.y() * this->w())))
           return false;
      if ( (sign(p.w()) != sign(this->w())) && 
         ( (sign(p.x()) != sign(this->x())) ||
           (p.x() == 0 && this->x() == 0)) && 
           ((sign(p.y()) != sign(this->y())) ||  
           (p.y() == 0 && this->y() == 0)) )
           return false;
      return true;
    }
    else {
       return (((p.y() * this->x()) == (p.x() * this->y()))) &&	
              (sign(p.x()) == sign(this->x())) &&
              (sign(p.y()) == sign(this->y()));
    }
	 
  }


  bool operator!=(const Point_T2 &p) const 
  {
    return !(*this == p);
  }

  bool antipodal(const Point_T2 &p) const 
  {
    Point_T2 q(p.w()*(-1),p.x()*(-1),p.y()*(-1));
    return (*this == q);
  }

  int dimension() const
  {
      return 2;
  }

  const 
  FT & cartesian(int i) const;

  Bbox_3 bbox() const;

  const  FT &
  operator[](int i) const;

  TPoint_T2 transform(const TAff_transformation_T2 &t){
    return t.transform(*this);
  }

 private:
   Point_3<rep> *Point;
};

template < class R >
inline 
const 
typename Point_T2<R>::FT & Point_T2<R>::cartesian(int i) const
{
  CGAL_kernel_precondition( (i>=0) && (i<=1) );
  CGAL_kernel_precondition_msg(w() != 0,"Point at Infinity (T^2)!");
  FT *c;
  if (i==0) 
    c = new FT(this->x()/this->w());
  else 
    c = new FT(this->y()/this->w()); 
  return *c;
}

template < class R >
Bbox_3
Point_T2<R>::bbox() const
{
  // FIXME: to_interval
  double bw = CGAL::to_double(this->w());
  double bx = CGAL::to_double(this->x());
  double by = CGAL::to_double(this->y());
  return CGAL::Bbox_3(bw, bx, by, bw, bx, by);
}

template < class R >
inline
const typename Point_T2<R>::FT &
Point_T2<R>::operator[](int i) const
{
  CGAL_kernel_precondition(i >=0 && i<=2);
  switch(i) {
    case 0:  return Point->x();
    case 1:  return Point->y();
    default: return Point->z();
  }
}

//#ifndef CGAL_CARTESIAN_NO_OSTREAM_INSERT_POINTC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Point_T2<R> &p)
{
    switch(os.iword(IO::mode)) {
    case IO::ASCII :
        return os << (p.w()) << ' ' << (p.x())  << ' ' << (p.y());
    case IO::BINARY :
        write(os, (p.w()));
        write(os, (p.x()));
        write(os, (p.y()));
        return os;
    default:
        os << "Point_T2(" << (p.w()) << ", " << (p.x())  << ", " << (p.y()) << ")";
        return os;
    }
}
//#endif // CGAL_CARTESIAN_NO_OSTREAM_INSERT_POINTC3


//#ifndef CGAL_CARTESIAN_NO_ISTREAM_EXTRACT_POINTC3
template < class R >
std::istream &
operator>>(std::istream &is, Point_T2<R> &p)
{
    typename R::FT x, y, z;
    switch(is.iword(IO::mode)) {
    case IO::ASCII :
        is >> x >> y >> z;
        break;
    case IO::BINARY :
        read(is, x);
        read(is, y);
        read(is, z);
        break;
    default:
        std::cerr << "" << std::endl;
        std::cerr << "Stream must be in ascii or binary mode" << std::endl;
        break;
    }
    if (is)
	p = Point_T2<R>(x, y, z);
    return is;
}
//#endif // CGAL_CARTESIAN_NO_ISTREAM_EXTRACT_POINTC3


CGAL_END_NAMESPACE

#endif //CGAL_POINT_T2_H
