// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Segment_2.h,v $
// $Revision: 1.14 $ $Date: 2003/10/21 12:18:13 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_SEGMENT_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_SEGMENT_T2_H

#include <CGAL/Point_T2.h>
#include <CGAL/Cartesian/Point_3.h>
#include <CGAL/Cartesian/Plane_3.h>
#include <CGAL/Cartesian/Segment_3.h>

CGAL_BEGIN_NAMESPACE

template < class R_ > 
class Segment_T2 {

  typedef R_                                    R;
  typedef typename R_::rep	                    rep;
  typedef typename R_::Point_T1	                TPoint_T1;
  typedef typename R_::Point_T2	                TPoint_T2;	
  typedef typename R_::Segment_T2               TSegment_T2;
  typedef typename R_::Vector_T2                TVector_T2;
  typedef typename R_::Direction_T2             TDirection_T2;
  typedef typename R_::Line_T2                  TLine_T2;
  typedef typename R_::Aff_transformation_T2    TAff_transformation_T2;

  protected:
    Segment_3<rep> *Segment;	
  
  public:
  
    Segment_T2(){}
   
    Segment_T2(const TPoint_T2 &sp, const TPoint_T2 &ep)
    {
      CGAL_kernel_precondition_msg(!(sp.antipodal(ep)), 
      " 'The' segment between p and q isn't defined. They are antipodal!!");
      Point_3<rep> p1(sp.w(),sp.x(),sp.y());
      Point_3<rep> p2(ep.w(),ep.x(),ep.y());
      Segment = new Segment_3<rep>(p1,p2);
    }
    

    bool has_on(const TPoint_T2 &p) const;
    bool collinear_has_on(const TPoint_T2 &p) const;

    bool operator==(const Segment_T2 &s) const;
    bool operator!=(const Segment_T2 &s) const;

    const TPoint_T2 &   source() const
    {
      Point_3<rep> p = Segment->source();
      TPoint_T2 *q = new TPoint_T2(p.x(),p.y(),p.z());
      return *q;
    }

    const TPoint_T2 &   target() const
    {
      Point_3<rep> p = Segment->target();		
      TPoint_T2 *q = new TPoint_T2(p.x(),p.y(),p.z());  
      return *q;
    }

    TLine_T2 supporting_line() const
    {
      return  TLine_T2(source(),target());
    }

    const TPoint_T2 &   start() const;

    const TPoint_T2 &   end() const;

    const TPoint_T2 &   min() const;

    const TPoint_T2 &   max() const;
  		
    const TPoint_T2 &   vertex(int i) const;
		
    const TPoint_T2 &   point(int i) const;

    const TPoint_T2 &   operator[](int i) const;

    const TDirection_T2 & direction() const;
		
    TSegment_T2   opposite() const;

    TPoint_T1 squared_length() const;
		
    const TVector_T2 to_vector();
  
    TSegment_T2 transform(const TAff_transformation_T2 &t)
    {
      TPoint_T2 p(source());
      TPoint_T2 q(target());
      return TSegment_T2(t.transform(p),t.transform(q));
    }


};


template < class R > 
bool Segment_T2<R>::operator==(const Segment_T2 &s) const
{
  TPoint_T2 p=this->source();
  TPoint_T2 q=s.source();
  if (p != q)
    return false;
  p=this->target();
  q=this->target();
  return (p==q);
}	

template < class R > 
bool Segment_T2<R>::operator!=(const Segment_T2 &s) const
{
  return !(*this==s);
}

template < class R >
const typename Segment_T2<R>::TPoint_T2 &  
Segment_T2<R>::start() const
{
  return this->source();
}

template < class R > 
const typename Segment_T2<R>::TPoint_T2 &  
Segment_T2<R>::end() const
{
  return this->end();
}

/*returns the point of s with lexicographically smallest coordinate*/
template < class R > 
const typename Segment_T2<R>::TPoint_T2 &  
Segment_T2<R>::min() const
{
  if (lexicographically_wxy_smaller(target(),source()))
    return target();
  return source();
}

/*returns the point of s with lexicographically largest coordinate*/
template < class R > 
const typename Segment_T2<R>::TPoint_T2 &  
Segment_T2<R>::max() const
{
  if (lexicographically_wxy_larger(source(),target()))
    return source();
  return target();
}

template < class R > 
const typename Segment_T2<R>::TDirection_T2 &  
Segment_T2<R>::direction() const
{
  TPoint_T2 a(source());
  TPoint_T2 b(target());
  CGAL_kernel_precondition(!is_at_infinity(a) || !is_at_infinity(b));
  TDirection_T2 *d;
  d = new TDirection_T2(b-a);
  return *d;
}

template < class R >
bool
Segment_T2<R>::has_on(const TPoint_T2 &p) const
{
  TPoint_T2 s(source()); 
  TPoint_T2 t(target());
  Point_3<Cartesian<typename R::FT> > a(s.w(),s.x(),s.y()); 
  Point_3<Cartesian<typename R::FT> > b(t.w(),t.x(),t.y());
  Point_3<Cartesian<typename R::FT> > c(p.w(),p.x(),p.y());
  Point_3<Cartesian<typename R::FT> > o(typename R::FT(0),typename R::FT(0),typename R::FT(0));
  Point_3<Cartesian<typename R::FT> > a1(-s.w(),-s.x(),-s.y()); 
  Line_3<Cartesian<typename R::FT> > l(a,a1);
  if (collinear(p,s,t)) {
    if (coplanar_orientation(c,o,a)!=coplanar_orientation(c,o,b)) {
      if (coplanar_orientation(a,a1,c) == 0)
        return (angle(a,o,c) == ACUTE);
      return (coplanar_orientation(a,a1,b)==coplanar_orientation(a,a1,c)); 
     }
    return (p==s && s==t);
  }
  return false;
}

template < class R > 
bool
Segment_T2<R>::collinear_has_on(const TPoint_T2 &p) const
{
  TPoint_T2 s(source()); 
  TPoint_T2 t(target());
  Point_3<Cartesian<typename R::FT> > a(s.w(),s.x(),s.y()); 
  Point_3<Cartesian<typename R::FT> > b(t.w(),t.x(),t.y());
  Point_3<Cartesian<typename R::FT> > c(p.w(),p.x(),p.y());
  Point_3<Cartesian<typename R::FT> > o(typename R::FT(0),typename R::FT(0),typename R::FT(0));
  Point_3<Cartesian<typename R::FT> > a1(-s.w(),-s.x(),-s.y()); 
  Line_3<Cartesian<typename R::FT> > l(a,a1);
  if (coplanar_orientation(c,o,a)!=coplanar_orientation(c,o,b)) {
    if (coplanar_orientation(a,a1,c) == 0)
      return (angle(a,o,c) == ACUTE);
    return (coplanar_orientation(a,a1,b)==coplanar_orientation(a,a1,c)); 
  }
  return (p==s && s==t);
}

template < class R > 
typename Segment_T2<R>::TPoint_T1 
Segment_T2<R>::squared_length() const
{
  CGAL_kernel_precondition(!CGAL::is_at_infinity(source()) &&
                           !CGAL::is_at_infinity(target()));
  return squared_distance(source(),target());
}

template < class R > 
typename Segment_T2<R>::TSegment_T2 
Segment_T2<R>::opposite() const
{
  TPoint_T2 p = this->source();
  TPoint_T2 q = this->target();
  return TSegment_T2(q,p);
}

template < class R > 
const typename Segment_T2<R>::TPoint_T2 &
Segment_T2<R>::point(int i) const
{
  Point_3<rep> p = Segment->point(i);
  
  TPoint_T2 *q = new TPoint_T2(p.x(),p.y(),p.z());
  return *q;
}

template < class R > 
const typename Segment_T2<R>::TPoint_T2 &
Segment_T2<R>::operator[](int i) const
{
  
  return vertex(i);
}

template < class R > 
const typename Segment_T2<R>::TPoint_T2 &
Segment_T2<R>::vertex(int i) const
{
  return point(i);
}

template < class R > 
const typename Segment_T2<R>::TVector_T2 
Segment_T2<R>::to_vector()
{
  CGAL_kernel_precondition(!is_at_infinity(source()) &&
                           !is_at_infinity(target()));
  return (target() - source()); 
}


//#ifndef CGAL_NO_OSTREAM_INSERT_SEGMENTC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Segment_T2<R> &s)
{
  Point_T2<R> p = s.source();
  Point_T2<R> q = s.target();	
  switch(os.iword(IO::mode)) {
  case IO::ASCII :
    return os << p << ' ' << q;
  case IO::BINARY :
    return os << p << q;
  default:
    return os << "Segment_T2(" << p <<  ", " << q << ")";
  }
}
//#endif // CGAL_NO_OSTREAM_INSERT_SEGMENTC3

//#ifndef CGAL_NO_ISTREAM_EXTRACT_SEGMENTC3
template < class R >
std::istream &
operator>>(std::istream &is, Segment_T2<R> &s)
{
  typename R::Point_T2 p, q;

  is >> p >> q;

  if (is)
    s = Segment_T2<R>(p, q);
  return is;
}
//#endif // CGAL_NO_ISTREAM_EXTRACT_SEGMENTC3

CGAL_END_NAMESPACE

#endif //CGAL_SIGNED_HOMOGENEOUS_SEGMENT_T2_H
 
