// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Sweep_line_2/include/CGAL/Sweep_line_2/Sweep_line_2_impl.h,v $
// $Revision: 1.8 $ $Date: 2003/09/18 10:25:48 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Tali Zvi <talizvi@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SWEEP_LINE_T2_IMPL_H
#define CGAL_SWEEP_LINE_T2_IMPL_H

#include <CGAL/Sweep_line_T2/Sweep_line_functors.h>
#include <CGAL/Sweep_line_T2/Event_info.h>
#include <CGAL/assertions.h>
#include <CGAL/functional.h>
#include <map>
#include <set>

#ifndef VERBOSE

#define SL_DEBUG(a)
#define PRINT_INSERT(a)
#define PRINT_ERASE(a)
#define PRINT_NEW_EVENT(p, e) 
#define DBG(a)
#define STORE_RESULT(a)

#else

#define SL_DEBUG(a) {a}
#define PRINT_INSERT(a) { std::cout << "+++ inserting "; \
                          (a)->Print(); \
                          std::cout << "    currentPos = " << m_currentPos \
                                    << "\n"; \
                          }
#define PRINT_ERASE(a)  { std::cout << "--- erasing " ; \
                          (a)->Print(); }
#define PRINT_NEW_EVENT(p, e) \
{ std::cout << "%%% a new event was created at " << (p) << std::endl; \
  (e)->Print(); }
#define DBG(a) { std::cout << a << std::endl; }
#define STORE_RESULT(a) {a}
#endif

#include <list>

CGAL_BEGIN_NAMESPACE

/*!
  Sweep_line_T2_impl is a class that implements the sweep line algorithm
  based on the algorithm of Bentley and Ottmann.
  It extends the algorithm to support not only segments but polylines and 
  general curves as well.
  The curves are defined by the traits class that is one of the teplate 
  arguments.

  The algorithm is also extended to support the following degenerate cases:
  - non x-monotone curves
  - vertical segments
  - multiple (more then two) segments intersecting at one point
  - curves beginning and ending on other curves.
  - overlapping curves

  There are two main functionalities supported by this algorithm:
  1. calculate the non intersecting curves that are a product of 
     intersecting a set of input curves
  2. calculate all the intersection points between the curves specified

  An extension of this algorithm is used to produce a planar map containing
  the input curves efficiently. \sa Pmwx_aggregate_insert.

  General flow:
  After the initialization stage, the events are handled from left to right.

  For each event

    First pass - handles special cases in which curves start or end 
                 at the interior of another curve
    Handle vertical curve (bottom) - special caring if the event 
                 is a bottom end of a vertical curve
    Handle vertical overlapping curves - special caring if there are
                 overlapping vertical curves passing through the event.
    Handle left curves - iterate over the curves that intersect 
                 at the event point and defined to the left of the 
                 event. This is mostly where the output (sub curves 
		 or intersection points) is produced.
    Handle vertical curve (top) - special caring if the event id 
                 the upper end of a vertical curve. In case of vertical 
		 curves, this is where the output is produced.
    Handle right curves - iterate over the curves that intersect 
                 the event point and defined to the right of the 
		 event point. This is where new intersection points 
		 are calculated.
  End

  Convensions through out the code:
  In order to make the code as readable as possible, some convensions were 
  made in regards to variable naming:

    xp - is the intersection point between two curves
    slIter - an iterator to the status line, always points to a curve.
    out - always refer to the output iterator

*/

template <class CurveInputIterator,  class SweepLineTraits_T2,
          class SweepEvent, class CurveWrap>
class Sweep_line_T2_impl
{
public:
  typedef SweepLineTraits_T2 Traits;
  typedef typename Traits::Point_T2 Point_T2;
  typedef typename Traits::Point_R3 Point_R3;
  typedef typename Traits::Curve_T2 Curve_T2;
  typedef typename Traits::Segment_T2 Segment;
  typedef typename Traits::X_monotone_curve_T2 X_monotone_curve_T2;
  typedef typename Point_T2::FT FT;

  typedef SweepEvent Event;
  typedef Point_less_functor<Point_T2, Traits> PointLess;
  typedef std::map<const Point_T2*, Event*, PointLess> EventQueue; 
  typedef typename EventQueue::iterator EventQueueIter;
  typedef typename EventQueue::value_type EventQueueValueType;
  typedef std::vector<Event*> EventPtrContainer;
  typedef typename EventPtrContainer::iterator EventPtrContainerIter;

  typedef typename Event::SubCurveIter EventCurveIter;

  typedef CurveWrap Subcurve;
  typedef std::list<Subcurve*> SubCurveList;
  typedef typename SubCurveList::iterator SubCurveListIter;

  typedef Status_line_curve_less_functor<Traits, Subcurve> StatusLineCurveLess;
  typedef typename StatusLineCurveLess::Compare_param CompareParams;
  typedef std::set<Subcurve*, StatusLineCurveLess> StatusLine;
  typedef typename StatusLine::iterator StatusLineIter;
  typedef typename Event::OverlapSweepCurveList OverlapSweepCurveList;
  typedef typename Event::OverlapSweepCurveListIter  OverlapSweepCurveListIter;

  typedef std::list<Event *> EventList;
  typedef typename EventList::iterator EventListIter;

  typedef std::list<X_monotone_curve_T2> CurveList;
  typedef typename CurveList::iterator CurveListIter;

  class  SweepLineGetSubCurves {};
  class  SweepLineGetPoints {};
  class  SweepLineGetInterCurveList {};
  class  SweepLinePlanarmap {};

  Sweep_line_T2_impl()  : m_traits(new Traits()), m_traitsOwner(true),
    m_includeEndPoints(true), m_found_intersection(false), is_first_point(true)
  {
  }
  Sweep_line_T2_impl(Traits *t) : m_traits(t), m_traitsOwner(false),
    m_includeEndPoints(true), m_found_intersection(false), is_first_point(true)
  {
  }

  virtual ~Sweep_line_T2_impl();

  /*!
   *  Given a range of curves, this function returns a list of points 
   *  that are the intersection points of the curves.
   *  The intersections are calculated using the sweep algorithm.
   *  \param curves_begin the input iterator that points to the first curve 
   *                      in the range.
   *  \param curves_end the input past-the-end iterator of the range.
   *  \param subcurves an iterator to the first curve in the range
   *                   of curves created by intersecting the input curves.
   *  \param endpoints if true, the end points of the curves are reported
   *                   as intersection points. Defaults to true.
   *  \param overlapping indicates whether there are overlapping curves
   *                     in the input range. Defaults to false.
   */
  template <class OutpoutIterator>
  void  get_intersection_points(CurveInputIterator begin, 
                                CurveInputIterator end, 
                                OutpoutIterator points,
                                bool includeEndPoints = true)
  { 
    init(begin, end);
    SL_DEBUG(
    )
    m_includeEndPoints = includeEndPoints;
    m_found_intersection = false;
    sweep(points, SweepLineGetPoints());
  }


protected:


  void init(CurveInputIterator begin, CurveInputIterator end);
  void init_curve(X_monotone_curve_T2 &curve);
  void init_curve_aux(X_monotone_curve_T2 &curve,
                      X_monotone_curve_T2 &curve_base,
                      bool original_curve);


  /*! The main loop to calculate intersections among the curves
   *  Looping over the events in the queue, for each event we first
   *  handle the curves that are tothe left of the event point (i.e., 
   *  curves that we are done with), and then we look at the curves 
   *  to the right of the point, which means we attept to find intersections
   *  between them and their neighbours on the sweep line.
   */

  template <class OutpoutIterator, class Op>
  void sweep(OutpoutIterator out, Op tag, bool stop_at_first_int=false)
  {
    EventQueueIter eventIter = m_queue->begin();
    m_prevPos = *eventIter->first;
    m_sweepLinePos = *(eventIter->first);
    typedef typename Traits::Kernel Kernel;
 
    especial_case_oscurve = false;
    
    while ( eventIter != m_queue->end() )
    {
      const Point_T2 *p = (eventIter->first);
      const Point_T2 o (FT(1),FT(0),FT(0));
      Point_R3 p_R3(p->w(),p->x(),p->y());
      Point_R3 o_R3(o.w(),o.x(),o.y());
      Point_R3 oa_R3(-o.w(),o.x(),o.y());
      Point_R3 m_sweepLinePos_R3(m_sweepLinePos.w(),m_sweepLinePos.x(),m_sweepLinePos.y());
      if (m_traits->orientation(o, m_sweepLinePos, *p) != 0) { 
        m_prevPos = m_sweepLinePos;
        m_overlapSweep.clear();
      }
      else if (int(coplanar_orientation(p_R3,o_R3,oa_R3))*int(coplanar_orientation(
                   m_sweepLinePos_R3,o_R3,oa_R3)) == -1) {
        m_prevPos = m_sweepLinePos;
        m_overlapSweep.clear();
      }
      m_sweepLinePos = *p;
      Point_T2 sweepPos = *p;

      if ((sweepPos == o) || (sweepPos == m_traits->antipode(o)))
        sweepPos = Point_T2(FT(0),FT(1),FT(0));
      
      for (StatusLineIter iter = m_statusLine->begin(); 
                          iter != m_statusLine->end(); iter++)
      {
        (*iter)->updateSweepLinePos(sweepPos);
      }
 
      m_currentPos = *p;
         
      m_currentEvent = eventIter->second;
      SL_DEBUG(std::cout << "------------- " << *p << " --------------"<< std::endl;
                   m_currentEvent->Print();)
      
      if ( m_traits->orientation(o,*(eventIter->first), m_sweepLinePos) != 0) {
        SL_DEBUG(std::cout << "====================== clearing miniq " 
                           << eventIter->first  << " " << m_prevPos << "\n";)
        m_miniq.clear();
      }
      m_miniq.push_back(m_currentEvent);
      handle_overlapSweepLine_curve_bottom(tag);
      handle_overlapSweepLine_overlap_curves();
      handle_left_curves(out, tag, 1, m_queue);
      
      m_queue->erase(eventIter);
    
      handle_overlapSweepLine_curve_top(out, tag);
      handle_right_curves(out, tag, m_queue);
   
      if (*p != o && *p!= antipode(o)) {
        if (finished_events.size() == 0) 
          finished_events.push_back(*p);
        else {
          Segment seg0 = Segment(o,finished_events[finished_events.size()-1]);
          Segment seg1 = Segment(antipode(o),finished_events[finished_events.size()-1]); 
          if (seg0.has_on(*p) || seg1.has_on(*p)) {
            if (m_currentEvent->does_contain_overlapSweep_curve()) {
              OverlapSweepCurveList &oscurves = m_currentEvent->get_overlapSweep_curves();
              OverlapSweepCurveListIter osciter = oscurves.begin();
              bool found = false;
              while (osciter != oscurves.end() && !found)
              {
                Subcurve *oscurve = *osciter;
                if (oscurve->get_first_end() == o)
                  found = true;
                else if (oscurve->get_second_end() == o)
                  found = true;
                osciter++;
              }
              if (found) {
                //report the intersection points
                for (unsigned int k=0; k < finished_events.size(); k++)
                  add_point_to_output(finished_events[k], out);
                finished_events.erase(finished_events.begin(),finished_events.end());
              }
            }
            finished_events.push_back(*p); 
          }
          else {
            finished_events.erase(finished_events.begin(),finished_events.end());
            finished_events.push_back(*p);
          }    
        }
        bool found = false;
        OverlapSweepCurveListIter osciter;
        if (m_currentEvent->does_contain_overlapSweep_curve()) {
          OverlapSweepCurveList &oscurves = m_currentEvent->get_overlapSweep_curves();
          osciter= oscurves.begin();
          while (osciter != oscurves.end() && !found)
          {
            Subcurve *oscurve = *osciter;
            if (oscurve->get_first_end() == antipode(o))
              found = true;
            else if (oscurve->get_second_end() == antipode(o))
              found = true;
            else
              osciter++;
          }
        }
        if (!especial_case_oscurve && found) {
            especial_event = *p;
            especial_case_oscurve = true;
        } 
        if (especial_case_oscurve) {
          Segment seg0 = Segment(o,especial_event);
          Segment seg1 = Segment(antipode(o),especial_event); 
          if (especial_event != *p) {
            if (seg0.has_on(*p) || seg1.has_on(*p))
              add_point_to_output(*p, out);
            else
              especial_case_oscurve = false;
          }
        }
      }
       
      eventIter = m_queue->begin(); 

      
      
      
    }
  }


  template <class OutpoutIterator>
  void handle_overlapSweepLine_curve_top(OutpoutIterator out, SweepLineGetPoints &tag)
  {
    SL_DEBUG(std::cout<<"handle_overlapSweepLine_curve_top... ";)
    if ( !m_currentEvent->does_contain_overlapSweep_curve() )
    {
      SL_DEBUG(std::cout<<"exiting\n ";)
      return;
    }
    SL_DEBUG(std::cout<<"\n ";)

    OverlapSweepCurveList &oscurves = m_currentEvent->get_overlapSweep_curves();
    OverlapSweepCurveListIter osciter = oscurves.begin();

    while ( osciter != oscurves.end() )
    {
      Subcurve *oscurve = *osciter; 
      Point_T2 topPoint = m_currentEvent->get_point();
      if ( oscurve->is_bottom_end(topPoint)) {
        SL_DEBUG(std::cout<<"this is the bottom. skipping.\n";)
        ++osciter;
        continue;
      }
      SL_DEBUG(std::cout<<"handling top point of vertical curve\n";)
      StatusLineIter slIter = m_statusLine->lower_bound(oscurve);
      if ( slIter != m_statusLine->end() )
      {
        SL_DEBUG(std::cout<<"starting at curve \n";)
        SL_DEBUG((*slIter)->Print();)
        const Point_T2 &bottomPoint = oscurve->get_bottom_end();
        std::cout << (*slIter)->get_curve() << std::endl;
        while (slIter != m_statusLine->end() && m_traits->point_in_PerpendicularDirSweep_range(
               (*slIter)->get_curve(), topPoint) && 
                (m_traits->curve_compare_dirSweepLine_at_perpendicularDir(topPoint,
                           (*slIter)->get_curve()) != SMALLER ||
                (m_traits->curve_compare_dirSweepLine_at_perpendicularDir(topPoint,
                           (*slIter)->get_curve()) != LARGER && bottomPoint == Point_T2(-1,0,0))) &&
                m_traits->point_in_PerpendicularDirSweep_range((*slIter)->get_curve(), 
                                       bottomPoint) && 
                (m_traits->curve_compare_dirSweepLine_at_perpendicularDir(bottomPoint,
                           (*slIter)->get_curve()) != LARGER ||
                (m_traits->curve_compare_dirSweepLine_at_perpendicularDir(bottomPoint,
                           (*slIter)->get_curve()) != SMALLER && bottomPoint == Point_T2(-1,0,0))))
        {
          SL_DEBUG(std::cout<<"checking \n";)
          SL_DEBUG((*slIter)->Print();) 
          typedef typename Point_T2::FT FT;
          Point_T2 o(FT(1),FT(0),FT(0));
          if ( orientation(o,(*slIter)->get_first_end(),topPoint) == COLLINEAR)
          {
            if (!m_traits->curves_equal((*slIter)->get_baseCurve(),oscurve->get_baseCurve())) {
              m_currentEvent->add_overlapSweep_curve_x_point((*slIter)->get_first_end());
              // if this point was not an event, 
              //we need to report the point
              // test40/42
              if ( !m_includeEndPoints && !is_internal_x_point((*slIter)->get_first_end())) {
                SL_DEBUG(std::cout << "Reporting point (1): " << (*slIter)->get_first_end() << "\n";)
                add_point_to_output((*slIter)->get_first_end(), out);
                m_found_intersection = true;
              }
            }
          }
          else if ((*oscurve).get_bottom_end() == o || 
                  (*oscurve).get_bottom_end() == Point_T2(FT(-1),FT(0),FT(0))) {
             Point_T2 currentPoint = m_currentEvent->get_point();
                Point_T2 xp;
                bool res = m_traits->nearest_intersection_to_right(oscurve->get_curve(), 
                                                                   (*slIter)->get_curve(), 
                                                                   currentPoint, 
                                                                   xp, xp);
                res = 0;
                EventQueueIter eqi = m_queue->find(&xp); 
                Event *e = 0;
                if ( eqi == m_queue->end() )
                {
                   e = new Event(xp, m_traits); 
                  #ifndef NDEBUG
                  e->id = m_eventId++;
                  #endif
                  m_events.push_back(e);
                  e->add_curve_to_left(*slIter, m_sweepLinePos);
                  e->add_curve_to_right(*slIter);

                  PRINT_NEW_EVENT(xp, e);

                  m_queue->insert(EventQueueValueType(&(e->get_point()), e)); 
                } 
                else {
                  e = eqi->second;
                  
                  e->mark_internal_intersection_point();
                  if (oscurve->get_bottom_end() != antipode(o)) 
                    e->add_curve(oscurve); // test41    
                  e->add_curve_to_left(*slIter, (*slIter)->get_first_end());
                  typedef typename Point_T2::FT FT;
                  Point_T2 o(FT(1),FT(0),FT(0));
                  if (orientation(o, m_currentEvent->get_point(),
                        (*slIter)->get_second_end()) == 1)
                    e->add_curve_to_right(*slIter);
                }
      
    
              }    
            
          ++slIter;
        }   
      }
      ++osciter;
    }
  }


/*!
 * Handles the degenerate case of vertical curves. Most of the cases
 * that occur with vertical curves are handled by this method and the
 * handle_vertical_curve_top method.
 *
 * When the current event is the bottom end of a vertical curve, we look
 * for intersection points between the vertical curve and any curve
 * in the status line that in the y-range that is defined by the bottom 
 * and top ends of the vertical curve. When those are found, we create
 * new events, unless ones already exist, in which case we update the events.
 * 
 * @param tag a tag that indicates the version of this method
 * \sa handle_vertical_curve_top
 */
 void handle_overlapSweepLine_curve_bottom(SweepLineGetPoints &tag)
{
  SL_DEBUG(std::cout<<"handle_vertical_curve_bottom... ";)

  Point_T2 currentPoint = m_currentEvent->get_point();
  if ( !m_currentEvent->does_contain_overlapSweep_curve() &&
       (currentPoint == Point_T2(-1,0,0) ||
        currentPoint == Point_T2(1,0,0)) )
  {
    SL_DEBUG(std::cout<<"exiting\n ";)
    return;
  }
  SL_DEBUG(std::cout<<"\n ";)
  OverlapSweepCurveList &oscurves = m_currentEvent->get_overlapSweep_curves();
  OverlapSweepCurveListIter osciter = oscurves.begin();
  
  while ( osciter != oscurves.end() )
  { 
    Subcurve *oscurve = *osciter;
    oscurve->updateSweepLinePos(m_currentPos);
    if ( oscurve->is_top_end(currentPoint))
    {
      ++osciter;
      continue;
    }
    
    SL_DEBUG(std::cout<<"handling bottom point of vertical curve\n";);
    StatusLineIter slIter = m_statusLine->lower_bound(oscurve);
    if ( slIter == m_statusLine->end() ) {
      SL_DEBUG(std::cout<<"no curves intersecting. exiting\n";);
      ++osciter;
      continue;
    }    

    SL_DEBUG(std::cout<<"starting at curve \n";);
    SL_DEBUG((*slIter)->Print(););
    const Point_T2 &topEnd = oscurve->get_top_end();
    EventQueueIter topEndEventIter = m_queue->find(&topEnd);
    CGAL_assertion(topEndEventIter!=m_queue->end());
    Event *topEndEvent = topEndEventIter->second;

    while (slIter != m_statusLine->end() && (! m_traits->point_in_PerpendicularDirSweep_range(
           (*slIter)->get_curve(), topEnd) ||
           m_traits->curve_compare_dirSweepLine_at_perpendicularDir(
                                     topEnd,(*slIter)->get_curve()) !=SMALLER) &&
           (! m_traits->point_in_PerpendicularDirSweep_range(
                                     (*slIter)->get_curve(), currentPoint) ||
           m_traits->curve_compare_dirSweepLine_at_perpendicularDir
                 (currentPoint,(*slIter)->get_curve()) != LARGER))
    {
      SL_DEBUG(std::cout<<"intersecting with \n";)
      SL_DEBUG((*slIter)->Print();) 

      if ( handle_overlapSweep_curve_x_at_end(oscurve, *slIter, topEndEvent, tag))
      {
         ++slIter;
         continue;
      }
      Point_T2 xp;
      bool res = m_traits->nearest_intersection_to_right(oscurve->get_curve(), 
                                                         (*slIter)->get_curve(), 
                                                         currentPoint, 
                                                         xp, xp);
      SL_DEBUG(CGAL_assertion(res==true);)
      res = 0;
      EventQueueIter eqi = m_queue->find(&xp); 
      Event *e = 0;
      if ( eqi == m_queue->end() )
      {
        e = new Event(xp, m_traits); 
        #ifndef NDEBUG
          e->id = m_eventId++;
        #endif
        m_events.push_back(e);
        e->add_curve_to_left(*slIter, m_sweepLinePos);
        e->add_curve_to_right(*slIter);

        PRINT_NEW_EVENT(xp, e);
        m_queue->insert(EventQueueValueType(&(e->get_point()), e)); 
      } 
      else {
        e = eqi->second;
        e->mark_internal_intersection_point();
        SL_DEBUG(std::cout << "Updating event \n";)
        SL_DEBUG(e->Print();)
        e->add_curve(oscurve); // test41    
        e->add_curve_to_left(*slIter, (*slIter)->get_first_end());
        typedef typename Point_T2::FT FT;
        Point_T2 o(FT(1),FT(0),FT(0));
        if (orientation(o, m_currentEvent->get_point(),
                        (*slIter)->get_second_end()) == 1)
          e->add_curve_to_right(*slIter);
      }
      
      topEndEvent->add_overlapSweep_curve_x_point(xp);
      ++slIter;
    }    
    ++osciter;
  }
  SL_DEBUG(std::cout<<"Done Handling overlapSweep\n";)
}


/*!
 * Handles overlapping vertical curves. 
 * If the current event point does not contain vertical curves, nothing is
 * done here.
 * For the current event point, we go through the list of vertical curves 
 * defined in the same x coordinate (m_verticals). For each curve, we check 
 * if the event point is in the interior of the vertical curve. If so, 
 * the event is set to be an intersection point (between the two 
 * vertical curves). 
 * While going through the vertical curves, if we reach a curve that the 
 * event point is above the curve, we remove the curve from the list.
 * 
 * Finally, we go thorugh the vertical curves of the event. If the event 
 * point is the bottom end of a vertical curve, we add the vertical curve 
 * to the list of vertical curves (m_verticals).
 */
 void handle_overlapSweepLine_overlap_curves()
 {
   SL_DEBUG(std::cout<<"\nhandle_vertical_overlap_curves... (" 
                      << m_currentEvent->get_point() << ")";)
  
   if ( !m_currentEvent->does_contain_overlapSweep_curve() ) {
     SL_DEBUG(std::cout << "no vertical - exiting\n";)
     return;
   }
   SL_DEBUG(std::cout << "\n";)
   SL_DEBUG(PrintOverlapSweeps();)
   const Point_T2 &point = m_currentEvent->get_point();
   SubCurveListIter iter = m_overlapSweep.begin();
   while ( iter != m_overlapSweep.end() && point!=Point_T2(1,0,0) &&
                                           point!=Point_T2(-1,0,0) )
   {
     Subcurve *curve = *iter;
     if (curve->get_first_end() == curve->get_second_end()
         && (curve->get_first_end() == Point_T2(1,0,0) ||
             curve->get_first_end() == Point_T2(-1,0,0))) {
       iter++;
       continue;
     } 
     else if ((curve->get_first_end() == Point_T2(1,0,0) ||
         curve->get_first_end() == Point_T2(-1,0,0)) &&
         curve->get_second_end() != Point_T2(1,0,0) &&
         curve->get_second_end() != Point_T2(-1,0,0)) {
         Segment seg0 = Segment(Point_T2(1,0,0),curve->get_second_end());
         Segment seg1 = Segment(Point_T2(-1,0,0),curve->get_second_end());  
         if (!(seg0.has_on(point) || seg1.has_on(point))) {
           iter++;
           continue;
       }
     }
     else if (curve->get_second_end() == Point_T2(1,0,0) ||
         curve->get_second_end() == Point_T2(-1,0,0)) {
         
       Segment seg0 = Segment(Point_T2(1,0,0),curve->get_first_end());
       Segment seg1 = Segment(Point_T2(-1,0,0),curve->get_first_end());  
       if (!(seg0.has_on(point) || seg1.has_on(point))){
         iter++;
         continue;
       }
     }
     if (m_traits->point_in_PerpendicularDirSweep_range(
           curve->get_curve(), point) &&
         m_traits->curve_compare_dirSweepLine_at_perpendicularDir
                          (point, curve->get_curve()) == LARGER &&
         curve->get_bottom_end() != Point_T2(-1,0,0))
     {
       iter = m_overlapSweep.erase(iter);

     }
     else if (!curve->is_end_point(point)) {
       EventQueueIter eventIter = m_queue->find(&(curve->get_top_end()));
       CGAL_assertion(eventIter!=m_queue->end());
       (eventIter->second)->add_overlapSweep_curve_x_point(point, true);
       m_currentEvent->mark_internal_intersection_point();
       ++iter;
     }
     else {
       ++iter;
     }
   }

   OverlapSweepCurveList &oscurves = m_currentEvent->get_overlapSweep_curves();
   OverlapSweepCurveListIter osciter = oscurves.begin();
   while (osciter != oscurves.end())
   {
     Subcurve *oscurve = *osciter;
    
     if ( oscurve->is_bottom_end(point) ) {
       m_overlapSweep.push_back(oscurve);
     }
     ++osciter;
   }
 }




  template <class OutpoutIterator>
  void handle_right_curves(OutpoutIterator out, SweepLineGetPoints &tag, EventQueue *queue)
  {
    SL_DEBUG(std::cout << "Handling right curves (" ;);
    SL_DEBUG(std::cout << m_currentEvent->get_point() << ")\n";);
    int numRightCurves = m_currentEvent->get_num_right_curves();
    if ( numRightCurves == 0 ) {
      return;
    }
    
    m_currentPos = m_sweepLinePos;
    if ( numRightCurves == 1 )
    {
      SL_DEBUG(std::cout << " - beginning of curve " << std::endl;);

      SL_DEBUG(Subcurve *tmp1 = *(m_currentEvent->right_curves_begin());
      PRINT_INSERT(tmp1););
      (*m_currentEvent->right_curves_begin())->updateSweepLinePos(m_currentPos);
      StatusLineIter slIter = m_statusLine->insert(m_status_line_insert_hint, 
                                    *(m_currentEvent->right_curves_begin()));
      (*(m_currentEvent->right_curves_begin()))->set_hint(slIter); 
      m_status_line_insert_hint = slIter; ++m_status_line_insert_hint;
      SL_DEBUG(PrintStatusLine(););
      
      // if this is the only curve on the status line, 
      // nothing else to do
      if ( m_statusLine->size() == 1 )
        return;
    
      StatusLineIter prev;
      prev = slIter;
      StatusLineIter next = slIter;
      ++next;

      SubCurveList mylist;
      if ( slIter != m_statusLine->begin() )
      {
        --prev;
        if ( CurveStartsAtCurve(*slIter, *prev) && !m_includeEndPoints) {
          SL_DEBUG(std::cout << "Reporting point (4): " 
                             << (*slIter)->get_first_end() 
                             << "\n";);
          if (!m_traits->curves_equal((*slIter)->get_baseCurve(),
                                      (*prev)->get_baseCurve())){
            add_point_to_output((*slIter)->get_first_end(), out);
          }
          std::cout << (*slIter)->get_first_end();
          m_found_intersection = true;
        }
        StatusLineIter tmp = prev;
        mylist.push_back(*prev);
        while ( tmp != m_statusLine->begin() ) 
        {
          --tmp;
          if ( do_curves_overlap(*prev, *tmp) )
            mylist.push_back(*tmp);
          else 
            break;
        }
      }
      if ( next != m_statusLine->end() )
      { 
        if ( CurveStartsAtCurve(*slIter, *next)  &&  !m_includeEndPoints) {
        if (!m_traits->curves_equal((*slIter)->get_baseCurve(),
                                  (*next)->get_baseCurve())) {
        SL_DEBUG(std::cout << "Reporting point (5): " << (*slIter)->get_first_end() << "\n";);
        add_point_to_output((*slIter)->get_first_end(), out);
        m_found_intersection = true;
      }
    }
    StatusLineIter tmp = next;
    mylist.push_back(*next);
    ++tmp;
    while ( tmp != m_statusLine->end() ) 
    {
      if ( do_curves_overlap(*next, *tmp) )
      {
        mylist.push_back(*tmp);
        ++tmp;
      }
      else
        break;
    }
  }
  intersect_curve_group(*(m_currentEvent->right_curves_begin()),mylist,queue);
  } 
  else
  {
    //std::cout << "ENTROU ELSE\n";
    /* this block takes care of 
    //
    //           /
    //          /
    //       --------
    //          \
    //           \
    */
    int numLeftCurves = m_currentEvent->get_num_left_curves();
     if ( numLeftCurves == 0 ) {
      SL_DEBUG(std::cout << " - handling special case " << std::endl;);
      StatusLineIter slIter;
      EventCurveIter currentOne = m_currentEvent->right_curves_begin();
      while ( currentOne != m_currentEvent->right_curves_end() ) {
        slIter = m_statusLine->lower_bound(*currentOne);
        if ( slIter != m_statusLine->end() ) {
          Subcurve *c = *slIter;
          if ( CurveStartsAtCurve(*currentOne, c) && !m_includeEndPoints) {
            if (!m_traits->curves_equal((*currentOne)->get_baseCurve(),
                                        c->get_baseCurve())) {
              SL_DEBUG(std::cout << "Reporting point (8): "
                                 << (*currentOne)->get_first_end() << "\n";);
  
              add_point_to_output((*currentOne)->get_first_end(),out);
  
              m_found_intersection = true;
              break;
            }
          }
        }
        currentOne++;
      }
    }
    // end block ..
      
      
    SubCurveList mylist;
    SubCurveList prevlist;
    SubCurveList currentlist;
     
         
    SL_DEBUG(std::cout << " - intersection point " << std::endl;);
    EventCurveIter firstOne = m_currentEvent->right_curves_begin();
    EventCurveIter lastOne = m_currentEvent->right_curves_end(); 
    --lastOne;
    EventCurveIter rightCurveEnd = m_currentEvent->right_curves_end(); 
       
    PRINT_INSERT(*firstOne);
  
    (*firstOne)->updateSweepLinePos(m_currentPos);   
  
    StatusLineIter slIter = m_statusLine->insert(m_status_line_insert_hint, *firstOne);
  
    (*slIter)->updateSweepLinePos(m_currentPos);
  
    (*firstOne)->set_hint(slIter);

    SL_DEBUG(PrintStatusLine(););
    if ( slIter != m_statusLine->begin() )
    { 
      StatusLineIter prev = slIter; --prev;
      if ( CurveStartsAtCurve(*slIter, *prev) && !m_includeEndPoints) {
        if (!m_traits->curves_equal((*slIter)->get_baseCurve(),
                                    (*prev)->get_baseCurve())) {
          SL_DEBUG(std::cout << "Reporting point (6): "<< (*slIter)->get_first_end() << "\n";);

          add_point_to_output((*slIter)->get_first_end(), out);
          m_found_intersection = true;
        }
      }
      // find all curves that are overlapping with the prev curve
      StatusLineIter tmp = prev;
      prevlist.push_back(*prev);
      while ( tmp != m_statusLine->begin() ) 
      {
        --tmp;
        if ( do_curves_overlap(*prev, *tmp))
          prevlist.push_back(*tmp);
        else
          break;
      }
      intersect_curve_group(*slIter, prevlist,queue);
    }

    currentlist.push_back(*firstOne);

    EventCurveIter currentOne = firstOne; ++currentOne;

    EventCurveIter prevOne = firstOne;
    while ( currentOne != rightCurveEnd )
    {

      m_currentPos = m_sweepLinePos;
      PRINT_INSERT(*currentOne);
      ++slIter;
      (*currentOne)->updateSweepLinePos(m_currentPos);
      slIter = m_statusLine->insert(slIter, *currentOne);
      (*currentOne)->set_hint(slIter);
      SL_DEBUG(PrintStatusLine(););
      if ( do_curves_overlap(*currentOne, *prevOne))
      {
        intersect_curve_group(*currentOne, currentlist,queue);
        currentlist.push_back(*currentOne);
      }
      else {
        prevlist = currentlist;
        currentlist.clear();
        currentlist.push_back(*currentOne);
      }
      intersect_curve_group(*currentOne, prevlist,queue);
      prevOne = currentOne;
      ++currentOne;
    }
    m_status_line_insert_hint = slIter; ++m_status_line_insert_hint;
    
    lastOne = currentOne; --lastOne;
    m_currentPos = m_sweepLinePos;
    PRINT_INSERT(*lastOne);
  
    SL_DEBUG(PrintStatusLine(););
    StatusLineIter next = slIter; ++next;
    if ( next != m_statusLine->end() ) {
      if ( CurveStartsAtCurve(*slIter, *next)  && !m_includeEndPoints) {
        if (!m_traits->curves_equal((*slIter)->get_baseCurve(),(*next)->get_baseCurve())) {
          SL_DEBUG(std::cout << "Reporting point (7): " 
                             << (*slIter)->get_first_end() << "\n";);

          add_point_to_output((*slIter)->get_first_end(), out);
          m_found_intersection = true;
        }
      }
      intersect_curve_group(*next, currentlist,queue);
      StatusLineIter tmp = next; ++tmp;
      while ( tmp != m_statusLine->end() ) 
      {
        if ( do_curves_overlap(*next, *tmp))
        {
          intersect_curve_group(*tmp, currentlist, queue);
          ++tmp;
        }
        else
          break;
      }
    }
  }
}
  


  // utility methods 
  bool intersect(Subcurve *c1, Subcurve *c2, EventQueue *queue);
  void intersect_curve_group(Subcurve *c1, SubCurveList &mylist, 															 EventQueue *queue);


  // reverse = false ==> we check if the curve starts at the list of curves
  // reverse = true ==> we check if any of the curves in the list start at 
  // the single curve
  template <class OutpoutIterator>
  void intersect_curve_group(Subcurve *c1, SubCurveList &mylist,
                             OutpoutIterator out, EventQueue *queue, bool reverse=false)
  {
    m_tmpOut.clear();
    SL_DEBUG(std::cout << "intersect_curve_group (with out)\n";)
    SL_DEBUG(std::cout << "intersecting with " << mylist.size()
             << " curves\n";)
    SubCurveListIter i = mylist.begin();
    while ( i != mylist.end())
    {
      bool flag;
      if ( reverse ) {
      }
      intersect(c1, *i, queue);
      ++i;
    }
    
    for ( SubCurveListIter iter = m_tmpOut.begin() ; 
          iter != m_tmpOut.end(); ++iter) {
      add_curve_to_output((*iter)->get_last_subcurve(), *iter, out);
    }
  }

  void remove_curve_from_status_line(Subcurve *leftCurve, EventQueue *queue);

  bool is_internal_x_point(const Point_T2 &p);
  
 
  bool do_curves_overlap(Subcurve *c1, Subcurve *c2);
 

  /*! Adds a new curve to the output list. If the overlapping flag is false,
      each unique curve is reported only once.

      @param out an iterator to the output container
      @param cv the curve to be added
  */
  template <class OutpoutIterator>
  void add_curve_to_output(const X_monotone_curve_T2 &cv, Subcurve *curve, 
                           OutpoutIterator out)
  {
    static Subcurve *prevCurve = 0;
    static X_monotone_curve_T2 prevXCv;

    if ( m_overlapping ) {
      *out = cv;
      ++out;
    } else {
      if ( prevCurve && similar_curves(cv, prevXCv)) {
        SL_DEBUG(std::cout << " curve already reported... " << std::endl;)
        return;
      }
      prevCurve = curve;
      prevXCv = cv;
      *out = cv;
      ++out;
    }
  }

  template <class OutpoutIterator>
  void add_point_to_output(const Point_T2 p, OutpoutIterator out) {

    SL_DEBUG(std::cout << "Reporting point " << p;)
		std::cout << p << std::endl;
		if ( is_first_point ) {
      is_first_point = false;
      *out = p;
      m_lastReportedPoint = p;
      ++out;
      SL_DEBUG(std::cout << " YES - first time\n";)
    } else {
      if ( m_lastReportedPoint != p ) {
	*out = p;
	++out;
	m_lastReportedPoint = p;
	SL_DEBUG(std::cout << " YES\n";)
      } else {
	SL_DEBUG(std::cout << " NO\n";)
      }
    }
  }



	bool handle_overlapSweep_curve_x_at_end(Subcurve *vcurve,
																					Subcurve *curve, 
			  																	Event *topEndEvent,
																				  SweepLineGetPoints tag);

  //
  // a second implementation for the get_intersection_points functionality
  //

 
  /*! For each left-curve, if it is the "last" subcurve, i.e., the 
   * event point is the right-edge of the original curve, the 
   * last sub curve is created and added to the result. Otherwise
   * the curve is added as is to the result.
   */
  template <class OutpoutIterator>
  void handle_left_curves(OutpoutIterator out, 
                          SweepLineGetPoints &tag,
                          int tag_queue,
                          EventQueue *queue)
  {
    SL_DEBUG(std::cout << "Handling left curve" << std::endl;)
    SL_DEBUG(m_currentEvent->Print();)
    const Point_T2 &eventPoint = m_currentEvent->get_point();
    if ( !m_currentEvent->has_left_curves())
    {
      if (m_includeEndPoints || m_currentEvent->is_internal_intersection_point() ||
         (m_currentEvent->get_num_overlapSweep_curves() +
          m_currentEvent->get_num_right_curves())  > 1) {
        if (is_statusLineFirstPosition()) {
          if (!(m_currentEvent->get_num_original_curves() >=1 ||
                m_currentEvent->get_num_overlapSweep_curves() > 3))
            return;
        }
        SL_DEBUG(std::cout << "Reporting point (2): " << eventPoint << "\n";)

        add_point_to_output(eventPoint, out);

        m_found_intersection = true;
      }
      return;
    }
    if (!is_statusLineFirstPosition()) {
    // delete the curve from the status line
      EventCurveIter leftCurveIter = m_currentEvent->left_curves_begin();
      m_currentPos = m_prevPos;
      m_use_hint_for_erase = false;
      while ( leftCurveIter != m_currentEvent->left_curves_end() )
      {
        // before deleting check new neighbors 
        // that will become after deletion
        remove_curve_from_status_line(*leftCurveIter,queue);
        m_use_hint_for_erase = true;
        PRINT_ERASE((*leftCurveIter));
        m_currentPos = m_prevPos;
        Subcurve *leftCurve = *leftCurveIter; 
        leftCurve->set_last_point(eventPoint);
        ++leftCurveIter;
      }
    }
    if ( m_includeEndPoints || 
         m_currentEvent->is_internal_intersection_point() ||
         (m_currentEvent->get_num_left_curves() +
          m_currentEvent->get_num_overlapSweep_curves() +
          m_currentEvent->get_num_right_curves()) > 1) {
      if (is_statusLineFirstPosition()) { 
        if (!(m_currentEvent->get_num_original_curves() >=1 ||
             (m_currentEvent->get_num_left_curves() +
              m_currentEvent->get_num_overlapSweep_curves() +
              m_currentEvent->get_num_right_curves()) > 3))
          return;
      }
      SL_DEBUG(std::cout << "Reporting point (3): " << eventPoint << "\n";)

      add_point_to_output(eventPoint, out);

      m_found_intersection = true;
    }
  }

  
#ifndef NDEBUG
  void PrintEventQueue(EventQueue *queue);

  void PrintStatusLine();

#endif

protected:
  /*! a pointer to a traits object */
  Traits *m_traits;

  /*! an indication to whether the traits should be deleted in the distructor
   */
  bool m_traitsOwner;

	bool is_statusLineFirstPosition();

  /*! if false, overlapping subcurves are reported only one. 
    Otherwise, they are reported as many times as they appeard. */
  bool m_overlapping;

  /*! if true, end points are reported as intersection points */
  bool m_includeEndPoints;

  /*! this is true when get_intersection_points() is called */
  bool m_stop_at_first_int;

  /*! used to hold all event, processed or not, so that they can 
`     all be deleted at the end */
  EventPtrContainer m_events; 

  /*! the queue of events (intersection points) to handle */
  EventQueue *m_queue;

  /*! the queue of events (intersection points) to handle */
  EventQueue *ao_queue;


  bool especial_case_oscurve;

  Point_T2 especial_event;

  /*! the queue of events (intersection points) to handle */
  EventQueue *final_queue;

  std::vector<Point_T2> finished_events;
 
  /*! The subcurves, as created on the fly */
  SubCurveList m_subCurves;

  /*! The status line */
  StatusLine *m_statusLine;

  /*! a struct that holds params associated with the curve compare functor */
  CompareParams *m_comp_param;

  /*! A reference point that is used for comapring curves. It is used
      when inserting/erasing curves from the status line. */
  Point_T2 m_currentPos;

  /*! A reference point that is used for comapring curves */
  Point_T2 m_prevPos;

  /*! The current position (in X) of the sweep line */
  Point_T2 m_sweepLinePos;

  /*! if non x-monotone are specified, this hold the x-monotone 
    curves created when splitting them into x-monotone curves. */
  std::vector<X_monotone_curve_T2> m_xcurves;

  /*! a pointer to thecurrent event */
  Event *m_currentEvent;

  /*! a queue that holds all the events that have the same x coordinate as 
      the status line. */
  EventList m_miniq;

  /*! a list of vertical curves at the x coordinate of the current event 
      point.*/
  SubCurveList m_overlapSweep;
  CurveList m_overlapSweepSubCurves;

  /*! when an intersection point is found this is turned to true */
  bool m_found_intersection;

  /*! An iterator of the  status line that is used as a hint for inserts. */
  StatusLineIter m_status_line_insert_hint;

  /*! A an indication wether the hint can be used to erase from the status
    line */
  bool m_use_hint_for_erase;
  SubCurveList m_tmpOut;
  Point_T2 m_lastReportedPoint;
  bool is_first_point;

  /*! a counter the is used to assign unique ids to the curves. */
  int m_curveId;

#ifndef NDEBUG
  int m_eventId;
#endif

  bool CurveStartsAtCurve(Subcurve *one, Subcurve *two)
  {

    SL_DEBUG(std::cout << "CurveStartsAtCurve: \n";)
    SL_DEBUG(std::cout << one->get_curve() << "\n" << two->get_curve() << "\n";)
    if ( m_traits->point_equal(one->get_first_end(),two->get_first_end()))
      return false;

    if ( !m_traits->point_equal(one->get_first_end(),m_currentEvent->get_point()) )
      return false;
    return (m_traits->has_on(two->get_curve(), one->get_first_end()));
  }

};

template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
~Sweep_line_T2_impl() 
{
  if ( m_traitsOwner ) delete m_traits;

  for ( SubCurveListIter sci = m_subCurves.begin() ; 
        sci != m_subCurves.end() ; ++sci)
  {
    delete *sci;
  }

  for ( EventPtrContainerIter ei = m_events.begin();
        ei != m_events.end() ; ++ei)
  {
    delete *ei;
  }
  delete m_queue;
  delete m_statusLine;
  delete m_comp_param;
}

/*! initializes the data structures to work with:
 *  - x-monotonize the inf\put curves
 *  - for each end point of each curve create an event
 *  - initialize the event queue
 *  -
 */
template <class CurveInputIterator,  class SweepLineTraits_T2,
          class SweepEvent, class CurveWrap>
inline void 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
init(CurveInputIterator begin, CurveInputIterator end)
{
  PointLess pred(m_traits);
  m_queue = new EventQueue(pred);
  final_queue = new EventQueue(pred);
  m_comp_param = new CompareParams(m_traits);
  StatusLineCurveLess slcurveless(m_comp_param);
  m_statusLine = new StatusLine(slcurveless);
  m_status_line_insert_hint = m_statusLine->begin();

  ao_queue = new EventQueue(pred);
#ifndef NDEBUG
  m_eventId = 0;
#endif
  m_curveId = 0;

  int count = 0;
  CurveInputIterator iter;
  for ( iter = begin ; iter != end ; ++iter)
  {

    if ( m_traits->is_dirSweepLine_monotone(*iter) ) {

      init_curve(*iter);
    }
    else
    {

      std::list<X_monotone_curve_T2> xcurves;
      m_traits->curve_make_dirSweepLine_monotone (*iter,
                                      std::back_inserter(xcurves));
      SL_DEBUG(
      std::cout << "curve " << *iter << " was split into " 
                << xcurves.size() << " curves." << std::endl;
      )
      for (typename std::list<X_monotone_curve_T2>::iterator i =
             xcurves.begin();
            i != xcurves.end() ; ++i )
      {
        m_xcurves.push_back(*i);
        init_curve(m_xcurves[count]);
        count++;
      }
    }
  }
}



/*! Given an x-monotone curve, create events for each end (if 
 *  one doesn't exist already). 
 *  For each curve create a Subcurve instance.
 */
template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
init_curve(X_monotone_curve_T2 &curve)
{

  X_monotone_curve_T2 subcurve1,subcurve2;

  if (m_traits->subdivide_curve(curve,subcurve1,subcurve2)) { 
    init_curve_aux(subcurve1,curve,false);
    init_curve_aux(subcurve2,curve,false);

  }
  else {

    init_curve_aux(curve,curve,true);

  }
}


template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
init_curve_aux(X_monotone_curve_T2 &curve,
               X_monotone_curve_T2 &curve_base,
               bool original_curve)
{
  typedef typename Point_T2::FT FT;
  
  const Point_T2 *source = new Point_T2(m_traits->curve_source(curve));
  const Point_T2 *target = new Point_T2(m_traits->curve_target(curve));
  Event *e0 = 0;
  Subcurve *subCv = new Subcurve(m_curveId++, curve, curve_base,
                                 &m_currentPos, m_traits,original_curve);
  m_subCurves.push_back(subCv);


  
  bool in = false;
#ifndef NDEBUG
#endif
  //handle the source point
  EventQueueIter eventIter;
  
    eventIter = m_queue->find(source); 

    if ( eventIter != m_queue->end() ) {

      if (*(eventIter->first) == *source) {
        in = true;
        SL_DEBUG(std::cout << "event " << source << " already exists\n";)

        e0 = eventIter->second;
      }
    }
    if (!in) { 
      e0 = new Event(*source, m_traits);
      e0->id = m_eventId++;
      m_events.push_back(e0); 
      m_queue->insert(EventQueueValueType(source, e0)); 
    }
    e0->add_curve(subCv);
    PRINT_NEW_EVENT(source, e0);

    in = false;
    eventIter = m_queue->find(target); 
    if ( eventIter != m_queue->end() ) {
      if (*(eventIter->first) == *target) {
        in = true;
        SL_DEBUG(std::cout << "event " << target << " already exists\n";)

        e0 = eventIter->second;
      }
    }
    if (!in) {
      e0 = new Event(*target, m_traits); 
      e0->id = m_eventId++;
      m_events.push_back(e0);
      m_queue->insert(EventQueueValueType(target, e0));
    }
    e0->add_curve(subCv);
    PRINT_NEW_EVENT(target, e0);

}

/*!
 * Perform intersection between the specified curve and all curves in the 
 * given group of curves.
 */ 
template <class CurveInputIterator, class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
intersect_curve_group(Subcurve *c1, 
                      SubCurveList &mylist, 
                      EventQueue *queue)
{
  m_tmpOut.clear();
  SL_DEBUG(std::cout << "intersecting with " << mylist.size() << " curves\n";)
  SubCurveListIter i = mylist.begin();
  while ( i != mylist.end())
  {
    std::cout << c1->get_curve() << std::endl;
    std::cout << (*i)->get_curve() << std::endl;

    intersect(c1, *i, queue);
    ++i;
  }
}



/*!
 * When a curve is removed from the status line for good, its top and
 * bottom neighbors become neighbors. This method finds these cases and
 * looks for the itnersection point, if one exists.
 *
 * @param leftCurve a pointer to the curve that is about to be deleted
 * @return an iterator to the position where the curve will be removed from.
 */

template <class CurveInputIterator, class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
remove_curve_from_status_line(Subcurve *leftCurve,EventQueue *queue)
{
  SL_DEBUG(std::cout << "remove_curve_from_status_line\n";)
  SL_DEBUG(PrintStatusLine();)
  SL_DEBUG(leftCurve->Print();)

  StatusLineIter sliter;
  sliter = leftCurve->get_hint(); 

  m_status_line_insert_hint = sliter; ++m_status_line_insert_hint;

  if ( !leftCurve->is_end_point(m_currentEvent->get_point())) {

    m_statusLine->erase(sliter);
    SL_DEBUG(std::cout << "remove_curve_from_status_line Done\n";)

    return;
  }

  m_currentPos = m_prevPos;
  CGAL_assertion(sliter!=m_statusLine->end());
  StatusLineIter end = m_statusLine->end(); 

  if ( sliter != m_statusLine->begin() && 
       m_statusLine->begin()!=m_statusLine->end()) 
  {
    end--;
    if (sliter != end ) {
      SubCurveList mylist;
      StatusLineIter prev = sliter; --prev;
    
      // collect all curves that overlap with *prev
      StatusLineIter tmp = prev;
      mylist.push_back(*prev);
      while ( tmp != m_statusLine->begin() ) 
      {
        --tmp;
        if ( do_curves_overlap(*prev, *tmp))
          mylist.push_back(*tmp);
        else
          break;//                                                             sect *next with the the *prev curve and all overlaps
      }
      StatusLineIter next = sliter; ++next;
      tmp = next;
      intersect_curve_group(*tmp, mylist,queue);
      // if there are curves that overlap with the *next curve, intersect
      // them with the *prev curve and all overlaps
      ++tmp;
      while ( tmp != m_statusLine->end() ) 
      {
        if ( do_curves_overlap(*next, *tmp))
        {
          intersect_curve_group(*tmp, mylist,queue);
          ++tmp;
        }
        else 
          break;
      }
    }
  }

  
  m_statusLine->erase(sliter);


  SL_DEBUG(std::cout << "remove_curve_from_status_line Done\n";)

           std::cout << "m_statusLine.size() = "<< m_statusLine->size() << std::endl;
} 

/*!
 * Handles the case in which a curve ont he status line passes through
 * one of the end points of the vertical curve.
 *
 * @param vcurve the vertical curve we are dealing with
 * @param curve a cerve that intersects with the vertical curve
 * @param topEndEvent the event attached to the top end of the vertical curve
 * @param tag 
 * @return returns true if the curve passed through one of the ends of the 
 *              vertical curve. Returns false otherwise.
 */
template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline bool
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
handle_overlapSweep_curve_x_at_end(Subcurve *vcurve, Subcurve *curve, 
                                   Event *topEndEvent, SweepLineGetPoints tag)
{
  const Point_T2 &topEnd = vcurve->get_top_end();
  // handle a curve that goes through the top point of the overlap sweep curve

  if (m_traits->point_in_PerpendicularDirSweep_range((curve)->get_curve(), topEnd) &&
      m_traits->curve_compare_dirSweepLine_at_perpendicularDir(
                                             topEnd, (curve)->get_curve()) == EQUAL)
  {
    if ( !curve->is_end_point(topEnd)) {

      topEndEvent->mark_internal_intersection_point();
    }
    return true;
  } 
  // handle a curve that goes through the bottom point of the vertical curve
  if (m_traits->point_in_PerpendicularDirSweep_range((curve)->get_curve(),
      m_currentEvent->get_point()) &&
      m_traits->curve_compare_dirSweepLine_at_perpendicularDir
            (m_currentEvent->get_point(),(curve)->get_curve()) == EQUAL)
  {
    if ( !curve->is_end_point(m_currentEvent->get_point())) {

      m_currentEvent->mark_internal_intersection_point();
    }
    return true;
  }
  return false;
}


template <class CurveInputIterator, class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline bool 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
is_statusLineFirstPosition()
{
  typedef typename Point_T2::FT FT;
  const Point_T2 eventPoint = m_currentEvent->get_point();
  const Point_T2 o(FT(1),FT(0),FT(0));
  const Point_T2 start(FT(0),FT(1),FT(0));
  if (orientation(o,eventPoint,start) == 0) {
    Point_R3 eventPoint_R3(eventPoint.w(),eventPoint.x(),eventPoint.y());
    Point_R3 o_R3(o.w(),o.x(),o.y());
    Point_R3 oa_R3(-o_R3.x(),o_R3.y(),o_R3.z());
    Point_R3 start_R3(start.w(),start.x(),start.y());
    return (int(coplanar_orientation(eventPoint_R3,o_R3,oa_R3))*
            int(coplanar_orientation(start_R3,o_R3,oa_R3)) != -1);
  }
  return false;
}

/*! 
 * Finds intersection between two curves. 
 * If the two curves intersect, create a new event (or use the event that 
 * already exits in the intersection point) and insert the curves to the
 * event.
 * @param curve1 a pointer to the first curve
 * @param curve2 a pointer to the second curve
 * @return true if the two curves overlap.
*/
template <class CurveInputIterator, class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline bool 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
intersect(Subcurve *c1, Subcurve *c2, EventQueue *queue)
{
  SL_DEBUG(std::cout << "Looking for intersection between:\n\t";)
  SL_DEBUG(c1->Print();)
  SL_DEBUG(std::cout << "\t";)
  SL_DEBUG(c2->Print();)
  SL_DEBUG(std::cout << "\n";)
  SL_DEBUG(std::cout << "relative to " << m_currentEvent->get_point() << "\n";)

  if ( c1->getId() == c2->getId() ) {
    SL_DEBUG(std::cout << "same curve, returning....\n";)
    return false;
  }

  Subcurve *scv1 = c1;
  Subcurve *scv2 = c2;
  const X_monotone_curve_T2 &cv1 = scv1->get_curve();
  const X_monotone_curve_T2 &cv2 = scv2->get_curve();

  bool isOverlap = false;

  Point_T2 xp, xp1;
  if ( m_traits->nearest_intersection_to_right(cv1, cv2, 
                                               m_currentEvent->get_point(), 
                                               xp, xp1))
  {
    if ( !m_traits->point_equal(xp, xp1)) {
      Point_T2 o(1,0,0);
      if ( m_traits->orientation(o,xp,xp1) == 1 )
        xp = xp1;
      SL_DEBUG(std::cout << "overlap detected\n";)
      isOverlap = true;
    }

    SL_DEBUG(
      std::cout << " a new event is created between:\n\t";
      scv1->Print();
      std::cout << "\t";
      scv2->Print();
      std::cout << "\trelative to ("
                << m_sweepLinePos << ")\n\t at (" 
                << xp << ")" << std::endl;
    )

    // check to see if an event at this point already exists...
    EventQueueIter eqi = m_queue->find(&xp);
    Event *e = 0;
    if ( eqi == queue->end() )
    {
      e = new Event(xp, m_traits); 
#ifndef NDEBUG
      e->id = m_eventId++;
#endif
      m_events.push_back(e);
      

      e->add_curve_to_left(c1, m_sweepLinePos);
      e->add_curve_to_left(c2, m_sweepLinePos);
      
      e->add_curve_to_right(c1);
      e->add_curve_to_right(c2);
      
      PRINT_NEW_EVENT(xp, e);
      m_queue->insert(EventQueueValueType(&(e->get_point()), e));
      return isOverlap;
    } 
    else 
    {
      SL_DEBUG(std::cout << "event already exists,updating.. (" << xp <<")\n";)
      e = eqi->second;
      e->add_curve_to_left(c1, m_sweepLinePos);
      if ( !scv1->is_end_point(xp)) { 
        e->add_curve_to_right(c1);
      }
      e->add_curve_to_left(c2, m_sweepLinePos);
      if ( !scv2->is_end_point(xp) ) {
        e->add_curve_to_right(c2);
      }
      SL_DEBUG(e->Print();)
      
    }
    
    return isOverlap;
  } 
  SL_DEBUG(std::cout << "not found T2\n";)
  return isOverlap;
}



template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline bool
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
is_internal_x_point(const Point_T2 &p)
{
  EventListIter itt = m_miniq.begin();
  while ( itt != m_miniq.end() )
  {
    if ( m_traits->point_equal(p, (*itt)->get_point())) 
    {
      if ((*itt)->is_internal_intersection_point()) {
        return true;
      }
      (*itt)->mark_internal_intersection_point(); 
                                     // this is to handle cases: |/ .
      return false;                  // (test 50/51)             |\ .
    } 
    ++itt;
  }
  CGAL_assertion(0);
  return false;
}



template <class CurveInputIterator,  class SweepLineTraits_T2,
          class SweepEvent, class CurveWrap>
inline bool
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
do_curves_overlap(Subcurve *c1, Subcurve *c2)
{
  SL_DEBUG(std::cout << "do_curves_overlap " << m_sweepLinePos << "\n" 
                     << "\t" << c1->get_curve() << "\n"
                     << "\t" << c2->get_curve() << "\n";)

  if ( m_traits->curves_overlap(c1->get_curve(),c2->get_curve()) )
    return true;

  return false;
}



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//                 point implementation of the methods                    //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//                         DEBUG UTILITIES                                //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

#ifndef NDEBUG

template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
PrintEventQueue(EventQueue *queue)
{
  SL_DEBUG(std::cout << std::endl << "Event queue: " << std::endl;)

  EventQueueIter iter = queue->begin();
  while ( iter != queue->end() )
  {
    SL_DEBUG(std::cout << "Point (" << iter->first << ")" << std::endl;)
    Event *e = iter->second;
    e->Print();
    ++iter;
  }
  SL_DEBUG(std::cout << "--------------------------------" << std::endl;)
}


template <class CurveInputIterator,  class SweepLineTraits_T2,
         class SweepEvent, class CurveWrap>
inline void 
Sweep_line_T2_impl<CurveInputIterator,SweepLineTraits_T2,SweepEvent,CurveWrap>::
PrintStatusLine()
{
  if ( m_statusLine->size() == 0) {
    std::cout << std::endl << "Status line: empty" << std::endl;
    return;
  }
  std::cout << std::endl << "Status line: (" 
            << m_currentPos << ")" << std::endl;
  StatusLineIter iter = m_statusLine->begin();
  while ( iter != m_statusLine->end() )
  {
    (*iter)->Print();
    ++iter;
  }
  std::cout << "Status line - end" << std::endl;
}
#endif

CGAL_END_NAMESPACE

#endif
