// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Triangle_2.h,v $
// $Revision: 1.14 $ $Date: 2003/10/21 12:18:16 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGLE_T2_H
#define CGAL_TRIANGLE_T2_H

CGAL_BEGIN_NAMESPACE

template <class R_> class Triangle_T2
{
  typedef typename R_::FT                       FT;
  typedef typename R_::Point_T2                 TPoint_T2;
  typedef typename R_::Point_T1                 TPoint_T1;
  typedef typename R_::Vector_T2                TVector_T2;
  typedef typename R_::Triangle_T2              TTriangle_T2;
  typedef typename R_::Aff_transformation_T2    TAff_transformation_T2;
  typedef typename R_::rep		                  rep;  

private:
	Triangle_3<rep> *Triangle; 

public:
  typedef R_                                     R;

  Triangle_T2(){}

  Triangle_T2(const TPoint_T2 &p, const TPoint_T2 &q, const TPoint_T2 &r)
  {
	//not any two antipodal points and the antipoda of any point cannot be in the 
	// segment between the other two       
    CGAL_kernel_precondition_msg(!(p.antipodal(q)) && 
                                 !(q.antipodal(r)) &&
                                 !(p.antipodal(r)),
                 "The triangle isn't defined if two vertices are antipodal!!");
    CGAL_kernel_precondition_msg(!has_on_segment(antipode(p),q,r) &&
                             !has_on_segment(antipode(q),p,r) &&
                             !has_on_segment(antipode(r),p,q),
   "The antipode of a vertice can't be in the segment defined by the other two!!"); 
    Triangle = new Triangle_3<rep>(Point_3<rep>(p.w(),p.x(),p.y()),
                                       Point_3<rep>(q.w(),q.x(),q.y()),
                                       Point_3<rep>(r.w(),r.x(),r.y()));
  }

  bool operator==(const Triangle_T2 &t) const;
  bool operator!=(const Triangle_T2 &t) const;
  
  bool is_degenerate() const;
  
  const TPoint_T2 & vertex(int i) const;
  const TPoint_T2 & operator[](int i) const;

  
  TPoint_T1 area() const;

  Orientation    orientation() const;

  Oriented_side  oriented_side(const TPoint_T2 &p) const;

  Bounded_side   bounded_side(const TPoint_T2 &p) const;

  bool           has_on_boundary(const TPoint_T2 &p) const;

  bool           has_on_bounded_side(const TPoint_T2 &p) const;
  bool           has_on_unbounded_side(const TPoint_T2 &p) const;

  bool           has_on_positive_side(const TPoint_T2 &p) const;
  bool           has_on_negative_side(const TPoint_T2 &p) const;

  TTriangle_T2 transform(const TAff_transformation_T2 &t)
  {
    return TTriangle_T2(t.transform(vertex(0)),
                t.transform(vertex(1)),
                t.transform(vertex(2)));
  }

 
};


template <class R> 
bool Triangle_T2<R>::operator==(const Triangle_T2 &t) const
{
  int i;
  for(i=0; i<3; i++)
    if ( vertex(0) == t.vertex(i) )
       break;

  return (i<3) && vertex(1) == t.vertex(i+1) && vertex(2) == t.vertex(i+2);
}

template <class R> 
bool Triangle_T2<R>::operator!=(const Triangle_T2 &t) const
{
  return !(*this==t);
}

template < class R >
const typename Triangle_T2<R>::TPoint_T2 &
Triangle_T2<R>::vertex(int i) const
{
  Point_3<rep> p=(Triangle->vertex(i));
  TPoint_T2 *q=new TPoint_T2(p.x(),p.y(),p.z());
  return *q;
}

template < class R>
const typename Triangle_T2<R>::TPoint_T2 & 
Triangle_T2<R>::operator[](int i) const
{
  return vertex(i);
}

template <class R> 
bool Triangle_T2<R>::is_degenerate() const
{
  return collinear(vertex(0),vertex(1),vertex(2));
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
typename Triangle_T2<R>::TPoint_T1
Triangle_T2<R>::area() const
{
  typename R::Vector_T2 v1 = vertex(1)-vertex(0);
  typename R::Vector_T2 v2 = vertex(2)-vertex(0);
  return TPoint_T1(FT(2) * v1.w() * v2.w(),
                  (v1.x() * v2.y() - v1.y() * v2.x() - 
                   v1.w() * v2.y() + v1.y() * v2.w() +
                   v1.w() * v2.x() - v1.x() * v2.w()));
}

template < class R >
inline
Orientation
Triangle_T2<R>::orientation() const
{
  return CGAL::orientation(vertex(0), vertex(1), vertex(2));
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
Bounded_side
Triangle_T2<R>::
bounded_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  Orientation o1 = CGAL::orientation(vertex(0), vertex(1), p),
              o2 = CGAL::orientation(vertex(1), vertex(2), p),
              o3 = CGAL::orientation(vertex(2), vertex(3), p);

 
  if (o2 == o1 && o3 == o1 && o1 != COLLINEAR)
    return ON_BOUNDED_SIDE;
  return (o1 == COLLINEAR
          && has_on_segment(p, vertex(0), vertex(1))) ||
         (o2 == COLLINEAR
          && has_on_segment(p, vertex(1), vertex(2))) ||
         (o3 == COLLINEAR
          && has_on_segment(p, vertex(2), vertex(3)))  
        ? ON_BOUNDARY
        : ON_UNBOUNDED_SIDE;
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
Oriented_side
Triangle_T2<R>::
oriented_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  // depends on the orientation of the vertices
  Orientation o1 = CGAL::orientation(vertex(0), vertex(1), p),
              o2 = CGAL::orientation(vertex(1), vertex(2), p),
              o3 = CGAL::orientation(vertex(2), vertex(3), p),
              ot = CGAL::orientation(vertex(0), vertex(1), vertex(2));

  if (o1 == ot && o2 == ot && o3 == ot) // ot cannot be COLLINEAR
    return Oriented_side(ot);
  return
     (o1 == COLLINEAR
       && has_on_segment(p, vertex(0), vertex(1))) ||
     (o2 == COLLINEAR
       && has_on_segment(p, vertex(1), vertex(2))) ||
     (o3 == COLLINEAR
       && has_on_segment(p, vertex(2), vertex(3)))
     ? ON_ORIENTED_BOUNDARY
     : Oriented_side(-ot);
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
bool
Triangle_T2<R>::
has_on_bounded_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  return bounded_side(p) == ON_BOUNDED_SIDE;
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
bool
Triangle_T2<R>::
has_on_unbounded_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{

  return bounded_side(p) == ON_UNBOUNDED_SIDE;
}

template < class R >
inline
bool
Triangle_T2<R>::
has_on_boundary(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  return bounded_side(p) == ON_BOUNDARY;
}

template < class R >
inline
bool
Triangle_T2<R>::
has_on_negative_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_NEGATIVE_SIDE;
}

template < class R >
inline
bool
Triangle_T2<R>::
has_on_positive_side(const typename Triangle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_POSITIVE_SIDE;
}

//#ifndef CGAL_NO_OSTREAM_INSERT_TRIANGLEC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Triangle_T2<R> &t)
{
    switch(os.iword(IO::mode)) {
    case IO::ASCII :
        return os << t[0] << ' ' << t[1] << ' ' << t[2];
    case IO::BINARY :
        return os << t[0]  << t[1]  << t[2];
    default:
        os << "Triangle_T2(" << t[0] <<  ", " << t[1] <<   ", " << t[2] <<")";
        return os;
    }
}
//#endif // CGAL_NO_OSTREAM_INSERT_TRIANGLEC3

//#ifndef CGAL_NO_ISTREAM_EXTRACT_TRIANGLEC3
template < class R >
std::istream &
operator>>(std::istream &is, Triangle_T2<R> &t)
{
    typename R::Point_T2 p, q, r;

    is >> p >> q >> r;

    if (is)
	t = Triangle_T2<R>(p, q, r);
    return is;
}
//#endif // CGAL_NO_ISTREAM_EXTRACT_TRIANGLEC3


CGAL_END_NAMESPACE

#endif
