// Copyright (c) 1997   INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Triangulation_2_traits_3.h,v $
// $Revision: 1.12 $ $Date: 2003/09/18 10:26:09 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGULATION_TRAITS_T2_H
#define CGAL_TRIANGULATION_TRAITS_T2_H

#include <CGAL/predicates_on_points_T1.h>
#include <CGAL/T1_constructions_T2.h>
#include <CGAL/point_constructions_T2.h>

CGAL_BEGIN_NAMESPACE

template <class kernel>
class Triangulation_traits_T2 {
public:
  typedef kernel K;
  typedef typename kernel::Point_T2                          Point_2;
  typedef typename Point_2::FT                               FT;
  typedef typename kernel::Point_T1                          Point_T1;
  typedef typename kernel::Segment_T2                        Segment_2;
  typedef typename kernel::Triangle_T2                       Triangle_2;
  typedef typename kernel::Line_T2                           Line_2;
  typedef typename kernel::Orientation_T2                    Orientation_2;
  typedef typename kernel::Compare_x_T2                      Compare_x_2;
  typedef typename kernel::Compare_y_T2                      Compare_y_2;
  typedef typename kernel::X_equal_T2                        X_equal_T2;
  typedef typename kernel::Y_equal_T2                        Y_equal_T2;
  typedef typename kernel::Equal_T2                          Equal_T2;
  typedef typename kernel::Orientation_T2                    Orientation_T2;
  typedef typename kernel::Squared_distance_T2               Squared_distance_T2;
  typedef typename kernel::Compare_T1p                       Compare_T1p;
  typedef typename kernel::Construct_circumcenter_T2         Construct_circumcenter_T2;
  typedef typename kernel::Construct_squared_radius_T2       Construct_squared_radius_T2;
  typedef typename kernel::Construct_segment_T2              Construct_segment_T2;

  Segment_2 construct_segment(const Point_2 & p, const Point_2 & q) const
  {
    return Segment_2(p,q);
  }

  Comparison_result compare_T1p(const Point_T1 & p, const Point_T1 & q) const
  {
    CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
    if (p.x()*q.w() < p.w()*q.x())
      return SMALLER;
    if ( (p.w() == 0) && (q.w() == 0) ) {
      if ( sign(p.x()) == sign(q.x()) )
        return EQUAL;
    }
    if ((p.x()*q.w() == q.x()*p.w()) &&
       (! ((sign(p.x()) != sign(q.x())) &&
       (sign(p.w()) != sign(q.w())) )))
      return EQUAL;
    return LARGER;
  }

  Point_T1 compute_distance(const Point_2 & p, const Point_2 & q) const
  {
    //a and b are proper points
    if (p.w()*q.w() != 0)
      return Point_T1(sign(p.w()*q.w()) * CGAL_NTS square(p.w()*q.w()),
                     (CGAL_NTS square(p.x()*q.w()-q.x()*p.w()) +
                     CGAL_NTS square(p.y()*q.w()-q.y()*p.w())));
    return Point_T1(FT(0),FT(1));
  }

  Comparison_result compare_distance(const Point_2 &p1, const Point_2 &p2,
                                     const Point_2 &q1, const Point_2 &q2) const
  {
     if ((p1.w() == 0 || p2.w() == 0) && (q1.w()*q2.w())!= 0) {
       if (sign(q1.w())==sign(q2.w()))
         return LARGER;
       else
         return SMALLER;
     }
     if ((q1.w() == 0 || q2.w() == 0) && (p1.w()*p2.w())!= 0) {
       if (sign(p1.w())==sign(p2.w()))
         return SMALLER;
       return LARGER;
     }
     if ((p1.w() == 0 || p2.w() == 0) && (q1.w() == 0 || q2.w() == 0)) {
       FT norm_p1 = sqrt(p1.w()*p1.w()+p1.x()*p1.x()+p1.y()*p1.y());
       FT norm_p2 = sqrt(p2.w()*p2.w()+p2.x()*p2.x()+p2.y()*p2.y());
       FT norm_q1 = sqrt(q1.w()*q1.w()+q1.x()*q1.x()+q1.y()*q1.y());
       FT norm_q2 = sqrt(q2.w()*q2.w()+q2.x()*q2.x()+q2.y()*q2.y());
       Point_2 p11 = Point_2(p1.w()/norm_p1, p1.x()/norm_p1, p1.y()/norm_p1);
       Point_2 p12 = Point_2(p2.w()/norm_p2, p2.x()/norm_p2, p2.y()/norm_p2);
       Point_2 q11 = Point_2(q1.w()/norm_q1, q1.x()/norm_q1, q1.y()/norm_q1);
       Point_2 q12 = Point_2(q2.w()/norm_q2, q2.x()/norm_q2, q2.y()/norm_q2);
       FT length_seg1 = ((p12.w()-p11.w())*(p12.w()-p11.w()) +
                         (p12.x()-p11.x())*(p12.x()-p11.x()) +
                         (p12.y()-p11.y())*(p12.y()-p11.y()));
       FT length_seg2 = ((q12.w()-q11.w())*(q12.w()-q11.w()) +
                         (q12.x()-q11.x())*(q12.x()-q11.x()) +
                         (q12.y()-q11.y())*(q12.y()-q11.y()));
       if (abs(length_seg1) > abs(length_seg2)) return LARGER;
       if (abs(length_seg1) < abs(length_seg2)) return SMALLER;
       return EQUAL;
     }
     Point_T1 sd1 = squared_distance(p1,p2);
     Point_T1 sd2 = squared_distance(q1,q2);
     return compare_T1p(sd1,sd2);
  }



  Point_T1 squared_radius(const Point_2 & p, const Point_2 & q, const Point_2 & r) const
  {
    Point_2 c = circumcenter(p,q,r);
    return squared_distance(p,c);
  }

  Point_T1 construct_squared_distance(const Point_2 & p, const Point_2 & q) const
  {
    return squared_distance(p,q);
  }

  Point_2 construct_circumcenter(const Point_2 & p,
                                 const Point_2 & q, const Point_2 & r) const
  {
    Point_2 c = circumcenter(p,q,r);
    if (sign(p.w()) != sign(c.w()) && c.w()!=0)
      return antipode(c);
    return c;
  }

};



CGAL_END_NAMESPACE

#endif // CGAL_TRIANGULATION_TRAITS_T2_H
