// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Vector_2.h,v $
// $Revision: 1.16 $ $Date: 2003/10/21 12:18:17 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_VECTOR_T2_H
#define CGAL_VECTOR_T2_H


CGAL_BEGIN_NAMESPACE

template <class R_ > class Vector_T2
{
  typedef R_ 		              R;
  typedef typename R_::FT             FT;
  typedef typename R_::rep            rep;
  typedef typename R_::Point_T2	      TPoint_T2;
  typedef typename R_::Vector_T2      TVector_T2;
  typedef typename R_::Segment_T2     TSegment_T2;
  typedef typename R_::Direction_T2   TDirection_T2;
  typedef typename R_::Point_T1       TPoint_T1;
  typedef typename R::Aff_transformation_T2   TAff_transformation_T2;
	
  private:
    Vector_3<rep> *Vector; 

  public:
    Vector_T2() {}
     
   
    Vector_T2(const TPoint_T2 &p)
    {
      CGAL_kernel_precondition_msg((p.w()!=0),
      "'The' vector with infinite source and/or target isn't defined.");

      Vector = new Vector_3<rep>(p.w(),p.x(),p.y());
    }

    Vector_T2(const TPoint_T2 &a, const TPoint_T2 &b)
    {
      CGAL_kernel_precondition_msg(!(antipodal(a,b)), 
      "'The' vector with antipodal source and target isn't defined.");

      CGAL_kernel_precondition_msg(((a.w()!=0) && (b.w()!=0)),
      "'The' vector with infinite source and/or target isn't defined.");

      Vector = new Vector_3<rep>(b.w()*a.w(),b.x()*a.w()-a.x()*b.w(),
                                 b.y()*a.w()-a.y()*b.w());
    }

    Vector_T2(const TDirection_T2 &d)
    {
      CGAL_kernel_precondition_msg((d.dw()!=0),
      "'The' vector with infinite source and/or target isn't defined.");

      Vector = new Vector_3<rep>(d.dw(),d.dx(),d.dy());
    }

    //the point [w,x,y] cannot be the antipodal of the origin
    Vector_T2(const FT &w, const FT &x, const FT &y) 
    {
      Vector = new Vector_3<rep>(w,x,y);
    }

    const FT & w() const
    {
      return Vector->x();
    }
  
    const FT & x() const
    {
      return Vector->y();
    }
  
    const FT & y() const
    {
      return Vector->z();
    }

    bool operator==(const TVector_T2 &p) const;
    
    bool operator!=(const TVector_T2 &p) const;

    TVector_T2 operator-() const;

    int dimension() const
    {
      return 2;
    }
	
    TVector_T2 operator+(const TVector_T2 &v) const;

    TVector_T2 operator-(const TVector_T2 &v) const;

    TVector_T2 operator/(const FT &c) const;

    TPoint_T1 operator*(const TVector_T2 &w) const;

    TPoint_T1 squared_length() const;

    bool operator|(const TVector_T2 &v) const;

    const FT & signed_homogeneous(int i) const;

    const FT & operator[](int i) const;


    TVector_T2 transform(const TAff_transformation_T2 &t) const
    {
      return t.transform(*this);
    }


};


template < class R > 
bool Vector_T2<R>::operator==(const TVector_T2 &s) const
{
  TPoint_T2 *p = new TPoint_T2(this->w(),this->x(),this->y());
  TPoint_T2 *q = new TPoint_T2(s.w(),s.x(),s.y());
  return (*p==*q);
}

template < class R > 
bool Vector_T2<R>::operator!=(const TVector_T2 &s) const
{
  return !(*this==s);
}

template < class R > 
typename Vector_T2<R>::TVector_T2 
Vector_T2<R>::operator-() const
{
  return TVector_T2(this->w(),-this->x(),-this->y());
}

template < class R >
inline
const typename Vector_T2<R>::FT &
Vector_T2<R>::operator[](int i) const
{
  return signed_homogeneous(i);
}

template < class R >
inline
const typename Vector_T2<R>::FT &
Vector_T2<R>::signed_homogeneous(int i) const
{
  CGAL_kernel_precondition( (i>=0) && (i<3) );
  if (i==0) return w();
  if (i==1) return x();
  return y();
}

template < class R > 
bool Vector_T2<R>::operator|(const TVector_T2 &v) const
{
   TPoint_T2 *p1=new TPoint_T2(v.w(),v.x(),v.y());
   TPoint_T2 *p2=new TPoint_T2(this->w(),this->x(),this->y());
   return (*p1|*p2);
}

template < class R > 
typename Vector_T2<R>::TVector_T2 
Vector_T2<R>::operator-(const TVector_T2 &v) const
{
  CGAL_kernel_precondition_msg(((this->w()!=0) && (v.w()!=0)), 
      "'The' vector with infinite extremes isn't defined.");

  CGAL_kernel_precondition_msg(!(*this|v),
  "'The' vector with antipodal source and target isn't defined.");

  return TVector_T2(this->w()*v.w(),this->x()*v.w()-v.x()*this->w(),
                                this->y()*v.w()-v.y()*this->w());
  
}

template < class R >
typename Vector_T2<R>::TVector_T2 
Vector_T2<R>::operator+(const TVector_T2 &v) const
{
  if ((this->w() < 0) && (v.w() < 0)) 
    return TVector_T2(-this->w()*v.w(),-this->x()*v.w()-v.x()*this->w(),
                      -this->y()*v.w()-v.y()*this->w());
  return TVector_T2(this->w()*v.w(),this->x()*v.w()+v.x()*this->w(),
                    this->y()*v.w()+v.y()*this->w());

}


template < class R >
typename Vector_T2<R>::TVector_T2 
Vector_T2<R>::operator/(const FT &c) const
{
  return TVector_T2(this->w()*c,this->x(),this->y());
}


template < class R >
typename Vector_T2<R>::TPoint_T1 
Vector_T2<R>::operator*(const TVector_T2 &w) const
{
  return TPoint_T1(this->w()*w.w(),this->x()*w.x() + this->y()*w.y());

}

template < class R >
typename Vector_T2<R>::TPoint_T1 
Vector_T2<R>::squared_length() const
{
  return TPoint_T1(this->w(), CGAL_NTS square(this->x()) + CGAL_NTS square(this->y()));
} 


//#ifndef CGAL_NO_OSTREAM_INSERT_SEGMENTC3
template < class R >
std::ostream &
operator<<(std::ostream &os, const Vector_T2<R> &s)
{
    switch(os.iword(IO::mode)) {
    case IO::ASCII :
        return os << s.w() << " " << s.x() << " " << s.y();
    case IO::BINARY :
        return os << s.w() << " " << s.x() << " " << s.y() ;
    default:
        return os << "Vector_T2(" << s.w() << "," << s.x() << "," << s.y() << ")";
    }
}
//#endif // CGAL_NO_OSTREAM_INSERT_SEGMENTC3

//#ifndef CGAL_NO_ISTREAM_EXTRACT_SEGMENTC3
template < class R >
std::istream &
operator>>(std::istream &is, Vector_T2<R> &s)
{
    typename R::Point_T2 p, q;

    is >> p >> q;

    if (is)
	s = Vector_T2<R>(p, q);
    return is;
}
//#endif // CGAL_CARTESIAN_NO_ISTREAM_EXTRACT_VECTORC3


CGAL_END_NAMESPACE

#endif
