// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/distance_predicates_2.h,v $
// $Revision: 1.9 $ $Date: 2003/10/21 12:18:21 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_DISTANCE_PREDICATES_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_DISTANCE_PREDICATES_T2_H

#include <CGAL/predicates_T2/kernel_T1T2.h>

CGAL_BEGIN_NAMESPACE

template <class R >
inline
Comparison_result
compare_distance_to_point(const Point_T2<R>& p,
                          const Point_T2<R>& q,
                          const Point_T2<R>& r)
{
  if (is_at_infinity(p) || is_at_infinity(q)) {
    if (is_at_infinity(q) || is_at_infinity(r)) {
      typename R::FT normp = sqrt(p.w()*p.w() +
                              p.x()*p.x() +
                              p.y()*p.y());
      typename R::FT normq = sqrt(q.w()*q.w() +
                              q.x()*q.x() +
                              q.y()*q.y());
      if (normq*abs(p.w()*r.x()+p.x()*r.y()+p.y()*r.w() -
                        r.x()*p.y()-p.w()*r.y()-r.w()*p.x()) <
              normp*abs(q.w()*r.x()+q.x()*r.y()+q.y()*r.w() -
                        r.x()*q.y()-q.w()*r.y()-r.w()*q.x()))
        return SMALLER;
      if (normq*abs(p.w()*r.x()+p.x()*r.y()+p.y()*r.w() -
                        r.x()*p.y()-p.w()*r.y()-r.w()*p.x()) >
              normp*abs(q.w()*r.x()+q.x()*r.y()+q.y()*r.w() -
                        r.x()*q.y()-q.w()*r.y()-r.w()*q.x()))
        return LARGER;
      return EQUAL;
    }
  }
  if (is_at_infinity(q)) {
    if (r.w()*p.w() > 0)
      return SMALLER;
    return LARGER;
  }
  if (is_at_infinity(p)) {
    if (r.w()*q.w() < 0)
      return SMALLER;
    return LARGER;
  }
  if (is_at_infinity(r))
    return compare_relative_distance_to_omega(p,q,r);
  return compare_T1p(squared_distance(p,q),

                 squared_distance(p,r));
}

/* return SMALLER if p is closer to the improper point r than q;
   return LARGER if q is closer to r than p;
   return EQUAL, otherwise */
template < class R >
inline
Comparison_result
compare_relative_distance_to_omega(const Point_T2<R> &p,
                                   const Point_T2<R> &q,
                                   const Point_T2<R> &r)
{

  CGAL_kernel_precondition(is_at_infinity(r) &&
                           !is_at_infinity(p) &&
                           !is_at_infinity(q));
  Line_T2<R> *l0,*l;
  if (p.w() * q.w() > 0) {
    l0 = new Line_T2<R>(p,r);
    l = new Line_T2<R>(p,perpendicular_direction(*l0));
  }
  else {
    l0 = new Line_T2<R>(Point_T2<R>(-p.w(),p.x(),p.y()),r);
    l = new Line_T2<R>(Point_T2<R>(-p.w(),p.x(),p.y()),perpendicular_direction(*l0));
  }
  if (l->has_on_positive_side(q))
     return SMALLER;
  if (l->has_on_negative_side(q))
     return LARGER;
  return EQUAL;
}
 


template < class R >
inline
bool
has_smaller_relative_distance_to_omega(const Point_T2<R> &p,
                                       const Point_T2<R> &q,
                                       const Point_T2<R> &r)
{
  CGAL_kernel_precondition(is_at_infinity(r));
  return compare_relative_distance_to_omega(p,q,r) == SMALLER;
}

template < class R >
inline
bool
has_larger_relative_distance_to_omega(const Point_T2<R> &p,
                                      const Point_T2<R> &q,
                                      const Point_T2<R> &r)
{
  CGAL_kernel_precondition(is_at_infinity(r));
  return compare_relative_distance_to_omega(p,q,r) == LARGER;
}

template < class R >
inline
bool
has_equal_relative_distance_to_omega(const Point_T2<R> &p,
                                     const Point_T2<R> &q,
                                     const Point_T2<R> &r)
{
  CGAL_kernel_precondition(is_at_infinity(r));
  return compare_relative_distance_to_omega(p,q,r) == EQUAL;

}

/* returns LARGER  iff the signed distance of p and l is larger 
   than the signed distance of q  and l, SMALLER, iff it is smaller,
   and EQUAL iff both are equal. */
template < class R >
inline
Comparison_result
compare_signed_distance_to_line ( const Line_T2<R> &l,
                                  const Point_T2<R> &p,
                                  const Point_T2<R> &q)
{
  return cmp_signed_dist_to_lineT2(l.W(),l.X(),l.Y(),
                                                  p,q);
}

/*returns LARGER iff the signed distance of r and l is larger 
than the signed distance of s and l, SMALLER, iff it is smaller,
 and EQUAL iff both are equal, where l is the directed line 
through p and q.*/
template < class R >
inline
Comparison_result
compare_signed_distance_to_line ( const Point_T2<R> &p,
                                  const Point_T2<R> &q,
                                  const Point_T2<R> &r,
                                  const Point_T2<R> &s)
{
  return compare_signed_distance_to_line(Line_T2<R>(p,q),r,s);
}

template < class R >
inline
bool
has_smaller_signed_distance_to_line ( const Point_T2<R> &p,
                                  const Point_T2<R> &q,
                                  const Point_T2<R> &r,
                                  const Point_T2<R> &s)
{
  return 
		(compare_signed_distance_to_line(Line_T2<R>(p,q),r,s)) == SMALLER;
}

template < class R >
inline
bool
has_larger_signed_distance_to_line ( const Point_T2<R> &p,
                                  const Point_T2<R> &q,
                                  const Point_T2<R> &r,
                                  const Point_T2<R> &s)
{
  return 
		(compare_signed_distance_to_line(Line_T2<R>(p,q),r,s)) == LARGER;
}


CGAL_END_NAMESPACE

#endif //CGAL_SIGNED_HOMOGENEOUS_DISTANCE_PREDICATES_T2_H
