// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/point_constructions_3.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:14:30 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T2_H

#include <CGAL/Point_T2.h>
#include <CGAL/Constructions_T2/kernel_ftT2.h>
#include <CGAL/kernel_ftT2.h>

CGAL_BEGIN_NAMESPACE


template <class R>
CGAL_KERNEL_LARGE_INLINE
Point_T2<R>
origin()
{
	typedef typename R::FT FT;
  return Point_T2<R>(FT(1),FT(0),FT(0));
}

template <class R>
CGAL_KERNEL_LARGE_INLINE
Point_T2<R>
dual(const Line_T2<R> &l)
{
  return Point_T2<R>(l.W(),l.X(),l.Y());
}

template <class R>
CGAL_KERNEL_LARGE_INLINE
Point_T2<R>
to_infinity(const Vector_T2<R> &v)
{
  return Point_T2<R>(0,v.x(),v.y());
}

template <class R>
CGAL_KERNEL_LARGE_INLINE
Point_T2<R>
point_on_line(int i, const Line_T2<R> &l)
{
  typename R::FT w, x, y;
  if (is_at_infinity(l.point0()))
    return l.point0();
  point_on_lineT2(l.point0().w(), l.point0().x(), l.point0().y(),
                  l.direction().dw(), l.direction().dx(), l.direction().dy(),
                  i, w, x, y);
  return Point_T2<R>(w, x, y);
}

template <class R>
CGAL_KERNEL_LARGE_INLINE
Point_T2<R>
projection_line(const Point_T2<R> &p, const Line_T2<R> &l)
{
  if (is_at_infinity(p)) {
    if (is_infinity_line(l))
      return p;
    return Point_T2<R>(typename R::FT(0),
                       l.direction().dx(),
                       l.direction().dy());
  }
  else
    CGAL_kernel_precondition(!is_infinity_line(l));
  typename R::FT w, x, y;
  projection_lineT2(p.w(), p.x(), p.y(),
		                l.point0().w(), l.point0().x(), l.point0().y(),
                    l.direction().dw(), l.direction().dx(), 
                    l.direction().dy(),
                    w, x, y);
  return Point_T2<R>(w, x, y);
}

template < class R >
inline
Point_T2<R>
midpoint(const Point_T2<R> &p,
         const Point_T2<R> &q )
{
  typename R::FT w, x, y;
  CGAL::midpointT2(p.w(), p.x(), p.y(), q.w(), q.x(), q.y(), w, x, y);
  return Point_T2<R>(w, x, y);
}

template < class R >
inline
Point_T2<R>
antipode(const Point_T2<R> &p)
{
  return Point_T2<R>(-p.w(), -p.x(), -p.y());
}

template < class R >
inline
Point_T2<R>
circumcenter(const Point_T2<R> &p, const Point_T2<R> &q, const Point_T2<R> &r)
{
  typedef typename Point_T2<R>::FT FT;

  if (is_at_infinity(p) && is_at_infinity(q)) {
    FT norm_p = sqrt(p.x()*p.x() + p.y()*p.y());
    Point_T2<R> p1 = Point_T2<R>(FT(0),p.x()/norm_p,p.y()/norm_p);
    FT norm_q = sqrt(q.x()*q.x() + q.y()*q.y());
    Point_T2<R> q1 = Point_T2<R>(FT(0),q.x()/norm_q,q.y()/norm_q);
    return Point_T2<R>(FT(0), (p1.x() + q1.x())/2, (p1.y() + q1.y())/2);
  }
  if (is_at_infinity(p) && is_at_infinity(r)) {
    FT norm_p = sqrt(p.x()*p.x() + p.y()*p.y());
    Point_T2<R> p1 = Point_T2<R>(FT(0),p.x()/norm_p,p.y()/norm_p);
    FT norm_r = sqrt(r.x()*r.x() + r.y()*r.y());
    Point_T2<R> r1 = Point_T2<R>(FT(0),r.x()/norm_r,r.y()/norm_r);
    return Point_T2<R>(FT(0), (p1.x() + r1.x())/2, (p1.y() + r1.y())/2);
  }
  if (is_at_infinity(q) && is_at_infinity(r)) {
    FT norm_q = sqrt(q.x()*q.x() + q.y()*q.y());
    Point_T2<R> q1 = Point_T2<R>(FT(0),q.x()/norm_q,q.y()/norm_q);
    FT norm_r = sqrt(r.x()*r.x() + r.y()*r.y());
    Point_T2<R> r1 = Point_T2<R>(FT(0),r.x()/norm_r,r.y()/norm_r);
    return Point_T2<R>(FT(0), (q1.x() + r1.x())/2, (q1.y() + r1.y())/2);
  }
  if (is_at_infinity(p)) {
    Direction_T2<R> dt = perpendicular_direction(q,r);
    return Point_T2<R>(0,dt.dx(),dt.dy());
  }
  if (is_at_infinity(q)) {
    Direction_T2<R> dt = perpendicular_direction(r,p);
    return Point_T2<R>(0,dt.dx(),dt.dy());
  }
  if (is_at_infinity(r)) {
    Direction_T2<R> dt = perpendicular_direction(p,q);
    return Point_T2<R>(0,dt.dx(),dt.dy());
  }

  Direction_T2<R> d1,d2;
  Line_T2<R> l1,l2;
  bool got_bisector1 = false;
  bool got_bisector2 = false;
  FT p1w;
  FT p1x;
  FT p1y;
  if (p.w()*q.w() > 0) {

    d1 = perpendicular_direction(p,q);
    l1 = Line_T2<R> (midpoint(p,q),Point_T2<R>(FT(0),d1.dx(),d1.dy()));
    got_bisector1 = true;
  }
  else {

    if (p.w() * r.w() > 0) {

      d2 = perpendicular_direction(r,p);
      return Point_T2<R>(0,d2.dx(),d2.dy());
    }
    else {

      d2 = perpendicular_direction(q,r);
      return Point_T2<R>(0,d2.dx(),d2.dy());
    }
  }
  if (r.w()*q.w() > 0) {

    d2 = perpendicular_direction(r,q);
    l2 = Line_T2<R> (midpoint(r,q),Point_T2<R>(FT(0),d2.dx(),d2.dy()));
    got_bisector2 = true;
  }
  else {

    if (p.w() * q.w() > 0) {
      Direction_T2<R> d2 = perpendicular_direction(p,q);
      return Point_T2<R>(0,d2.dx(),d2.dy());
    }
    else {

      Direction_T2<R> d2 = perpendicular_direction(r,p);
      return Point_T2<R>(0,d2.dx(),d2.dy());
    }
  }
  if (got_bisector1 && got_bisector2) {

    p1w = l1.X()*l2.Y()-l2.X()*l1.Y();
    p1x = -l1.W()*l2.Y()+l2.W()*l1.Y();
    p1y = l1.W()*l2.X()-l2.W()*l1.X();
  }
  else {
    if (!got_bisector1) {

      d1 = perpendicular_direction(p,r);
      l1 = Line_T2<R>(midpoint(p,r),Point_T2<R>(FT(0),d1.dx(),d1.dy()));
    }
    else {
      
      d2 = perpendicular_direction(p,r);
      l2 = Line_T2<R>(midpoint(p,r),Point_T2<R>(FT(0),d2.dx(),d2.dy()));
    }
    
    p1w = l1.X()*l2.Y()-l2.X()*l1.Y();
    p1x = -l1.W()*l2.Y()+l2.W()*l1.Y();
    p1y = l1.W()*l2.X()-l2.W()*l1.X();
  }
  Point_T2<R> p1(p1w,p1x,p1y);
  return p1;
}

template < class R >
inline
Point_T2<R>
project_to_the_other_range(const Point_T2<R> &p)
{
	if (!is_at_infinity(p))
		return antipode(p);
	return p;
}

template < class R >
inline
Point_T2<R>
compute_greater_y_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q)) {
    if (p.y() > 0  && q.y() <= 0) 
      return p;
    if (p.y() == 0 && q.y() < 0)
      return p;
    return q;
  }
  if (is_at_infinity(p)) {
    if (p.y() > 0) 
      return p;
    if (p.y() == 0 && sign(q.w()*q.y()) < 0)
      return p;
    return q;
  }
  if (is_at_infinity(q)) {
    if (q.y() == 0 && sign(p.w()*p.y()) > 0)
      return p;
    if (q.y() < 0)
      return p;
    return q;
  }
  if (p.y()*abs(q.w())*sign(p.w()) > 
      abs(p.w())*q.y()*sign(q.w()))
    return p;
  return q;
}

template < class R >
inline
Point_T2<R>
compute_greater_x_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q)) {
    if (p.x() > 0  && q.x() <= 0)
      return p;
    if (p.x() == 0 && q.x() < 0)
      return p;
    return q;
  }
  if (is_at_infinity(p)) {
    if (p.x() > 0)
      return p;
    if (p.x() == 0 && sign(q.w()*q.x()) < 0)
      return p;
    return q;
  }
  if (is_at_infinity(q)) {
    if (q.x() == 0 && sign(p.w()*p.x()) > 0)
      return p;
    if (q.x() < 0)
      return p;
    return q;
  }
  if (p.x()*abs(q.w())*sign(p.w()) >
      abs(p.w())*q.x()*sign(q.w()))
    return p;
  return q;
}

template < class R >
inline
Point_T2<R>
compute_smaller_x_one_range_infinity_T2(const Point_T2<R> &p,
                                        const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q)) {
    if (p.x() < 0  && q.x() >= 0) 
      return p;
    if (p.x() == 0 && q.x() > 0)
      return p;
    return q;
  }
  if (is_at_infinity(p)) {
    if (p.x() < 0)
      return p;
    if (p.x() == 0 && sign(q.w()*q.x()) < 0)
      return p;
    return q;
  }
  if (is_at_infinity(q)) {
    if (q.x() == 0 && sign(p.w()*p.x()) < 0)
      return p; 
    if (q.x() > 0)
      return p;
    return q;
  }
  if (p.x()*abs(q.w())*sign(p.w()) < 
      abs(p.w())*q.x()*sign(q.w()))
   return p;
  return q;
}

template < class R >
inline
Point_T2<R>
compute_smaller_y_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q)) {
    if (p.y() < 0  && q.y() >= 0) 
      return p;
    if (p.y() == 0 && q.y() > 0)
      return p;
    return q;
  }
  if (is_at_infinity(p)) {
    if (p.y() < 0)
      return p;
    if (p.y() == 0 && sign(q.w()*q.y()) < 0)
      return p;
    return q;
  }
  if (is_at_infinity(q)) {
    if (q.y() == 0 && sign(p.w()*p.y()) < 0)
      return p; 
    if (q.y() > 0)
      return p;
    return q;
  }
  if (p.y()*abs(q.w())*sign(p.w()) < 
      abs(p.w())*q.y()*sign(q.w()))
    return p;
  return q;
}

CGAL_END_NAMESPACE

#endif //CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T2_H
