// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/predicates_on_points_2.h,v $
// $Revision: 1.11 $ $Date: 2003/10/21 12:18:26 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PREDICATES_ON_POINTS_T1
#define CGAL_PREDICATES_ON_POINTS_T1

#include <CGAL/Point_T1.h>


CGAL_BEGIN_NAMESPACE

template < class R >
inline
bool
is_at_infinity_T1(const Point_T1<R> &p)
{
  return (p.w() == 0);
}


template < class R >
inline
bool
is_at_back_range_T1(const Point_T1<R> &p, const Point_T1<R> &q)
{
  return (p.w() < 0) && (q.w() < 0);
}

template < class R >
inline
bool
is_at_back_range_T1(const Point_T1<R> &p)
{
  return (p.w() < 0);
}


template < class R >
inline
bool
is_at_front_range_T1(const Point_T1<R> &p)
{
  return p.w() > 0;
}

template < class R >
inline
bool
is_zero_T1(const Point_T1<R> &p)
{
  return (p.x() == 0);
}

template < class R >
inline
bool
antipodal_T1(const Point_T1<R> &p,const Point_T1<R> &q)
{
  if (p.w()*q.x() != q.w()*p.x())
    return false;
  return (sign(p.w())!=sign(q.w()) && (sign(p.x())!=sign(q.x()) ||
         (sign(p.x()) == 0 && sign(q.x()) == 0)) );
}


template < class R >
inline
bool
less_equal_T1p(const Point_T1<R> &p,const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
  return lessT1p(p,q) || equalT1(p,q);
}

/*returns TRUE iff p is greater than q, otherwise returns false;
Precondition p,q are points at the front range of T1 or
at infinity*/
template < class R >
inline
bool
is_greater_front_range_infinity_T1(const Point_T1<R> &p,
														 const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() >= 0 && q.w() >= 0);
  if ((p == Point_T1<R>(0,1)) || (q == Point_T1<R>(0,-1)))
    return p!=q;
  if ((p == Point_T1<R>(0,-1)) || (q == Point_T1<R>(0,1)))
    return false;
  return (p.x()*q.w() > p.w()*q.x());
}

template < class R >
inline
Comparison_result
compare_front_range_infinity_T1(const Point_T1<R> &p,
														 const Point_T1<R> &q)
{
	CGAL_kernel_precondition(p.w() >= 0 && q.w() >= 0);
  if (is_greater_front_range_infinity_T1(p,q))
		return LARGER;
	if (is_smaller_front_range_infinity_T1(p,q))
		return SMALLER;
	return EQUAL;
}

/* the suffix T1p means T1+ - a half part of T1 where x is positive*/
template < class R >
inline
bool
is_smaller_T1p(const Point_T1<R> &p,const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
  return (p.x()*q.w() < p.w()*q.x());
}

template < class R >
inline
bool
equal_T1(const Point_T1<R> &p,const Point_T1<R> &q)
{
  if ( (p.w() == 0) && (q.w() == 0) )
     return ( sign(p.x()) == sign(q.x()) );
  if (p.x()*q.w() != q.x()*p.w())
     return false; 
  return (! ( ( sign(p.x()) != sign(q.x()) ) && 
				 ( sign(p.w()) != sign(q.w()) ) ) );

}

/* the suffix T1p means T1+ - a half part of T1 where x is positive */
template < class R >
inline
Comparison_result compare_T1p(const Point_T1<R> &p, const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
  if (is_smaller_T1p(p,q)) 
     return SMALLER;
  return (equal_T1(p,q)) ? EQUAL : LARGER;
}

template < class R >
inline
bool is_smaller_cartesian(const Point_T1<R> &p,
											 const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() > 0 &&
													 q.w() > 0) ;
	return (p.x()*q.w() < p.w()*q.x());
}

template < class R >
inline
bool is_smaller_or_equal_cartesian(const Point_T1<R> &p, 
											 const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() > 0 && 
													 q.w() > 0) ;
	return (p.x()*q.w() <= p.w()*q.x());
}

template < class R >
inline
bool is_smaller_front_range_infinity_T1(
											 const Point_T1<R> &p, 
											 const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() >= 0 && q.w() >= 0);
	if (p.w() > 0 && q.w() > 0)
		return is_smaller_cartesian(p,q);
	return (p == Point_T1<R>(0,-1));
}

template < class R >
inline
bool
smaller_one_range_infinity_T1(const Point_T1<R> &p,
             const Point_T1<R> &q)
{
	CGAL_kernel_precondition(p.w()*q.w()>=0);

	if (is_at_infinity_T1(p) && is_at_infinity_T1(q))
		return (p.x() < 0  && q.x() >= 0) ||
					 (p.x() == 0 && q.x() > 0);
	if (is_at_infinity_T1(p))
		return (p.x() < 0) || (p.x() == 0 && sign(q.w()*q.x()) > 0);
	if (is_at_infinity_T1(q))
		return (q.x() == 0 && sign(p.w()*p.x()) < 0) || q.x() > 0;
	return (p.x()*q.w() < p.w()*q.x());
}
CGAL_END_NAMESPACE


#endif //CGAL_PREDICATES_ON_POINTS_T1










